import traceback, urllib3, json

from base64 import b64decode, b64encode
from urllib.parse import urlparse
from datetime import datetime

from . import logging, __version__
from .cache import TTLCache, LRUCache
from .misc import formatUTC


version = '.'.join([str(num) for num in __version__])


class httpClient:
	def __init__(self, pool=100, timeout=30, headers={}, agent=None):
		self.cache = LRUCache()
		self.pool = pool
		self.timeout = timeout
		self.agent = agent if agent else f'IzzyLib/{version}'
		self.headers = headers

		self.client = urllib3.PoolManager(num_pools=self.pool, timeout=self.timeout)
		self.headers['User-Agent'] = self.agent


	def _fetch(self, url, new_headers={}, method='GET', data=None, cached=True):
		cached_data = self.cache.fetch(url)
		headers = self.headers.copy()
		headers.update(new_headers)

		if cached and cached_data:
			logging.debug(f'Returning cached data for {url}')
			return cached_data

		if not headers.get('User-Agent'):
			headers.update({'User-Agent': self.agent})

		logging.debug(f'Fetching new data for {url}')

		try:
			if data:
				if isinstance(data, dict):
					data = json.dumps(data)

				resp = self.client.request(method, url, headers=headers, body=data)

			else:
				resp = self.client.request(method, url, headers=headers)

		except Exception as e:
			logging.debug(f'Failed to fetch url: {e}')
			return

		if cached:
			logging.debug(f'Caching {url}')
			self.cache.store(url, resp)

		return resp


	def raw(self, *args, **kwargs):
		'''
			Return a response object
		'''
		return self._fetch(*args, **kwargs)


	def text(self, *args, **kwargs):
		'''
			Return the body as text
		'''
		resp = self._fetch(*args, **kwargs)

		return resp.data.decode() if resp else None


	def json(self, *args, **kwargs):
		'''
			Return the body as a dict if it's json
		'''

		headers = kwargs.get('headers')

		if not headers:
			kwargs['headers'] = {}

		kwargs['headers'].update({'Accept': 'application/json'})
		resp = self._fetch(*args, **kwargs)

		try:
			data = json.loads(resp.data.decode())

		except Exception as e:
			logging.debug(f'Failed to load json: {e}')
			return

		return data
