import shutil, sqlite3, traceback

from os.path import isfile
from collections import OrderedDict
from sqlite3 import OperationalError
from contextlib import contextmanager
from pathlib import Path
from datetime import datetime

from ..Lib.IzzyLib.cache import LRUCache
from ..Lib.IzzyLib.misc import boolean
from ..Lib.IzzyLib import logging

from DBUtils.PooledDB import PooledDB

from .. import dbversion, prev_dbversion
from ..config import var
from ..functions import DotDict


tables = DotDict({
	'config': {
		'id': {'type': 'integer', 'options': ['primary key']},
		'key': {'type': 'text', 'options': ['unique']},
		'value': {'type': 'text'},
	},
	'bookmarks': {
		'id': {'type': 'integer', 'options': ['primary key']},
		'name': {'type': 'text'},
		'url': {'type': 'text', 'options': ['unique']},
		'description': {'type': 'text'},
		'category': {'type': 'text'},
		'lastupdate': {'type': 'datetime'},
	},
	'links': {
		'id': {'type': 'integer', 'options': ['primary key']},
		'linkid': {'type': 'integer'}
	},
	'mastodon': {
		'id': {'type': 'integer', 'options': ['primary key']},
		'username': {'type': 'text'},
		'displayname': {'type': 'text'},
		'domain': {'type': 'text'},
		'fullname': {'type': 'text', 'options': 'unique'},
		'apikey': {'type': 'text'},
		'tootlimit': {'type': 'integer'},
		'avatar': {'type': 'text'},
		'lastupdate': {'type': 'datetime'},
	},
	'siteoptions': {
		'id': {'type': 'integer', 'options': ['primary key']},
		'domain': {'type': 'text', 'options': 'unique'},
		'microphone': {'type': 'BOOLEAN', 'default': False},
		'notification': {'type': 'BOOLEAN', 'default': False},
		'camera': {'type': 'BOOLEAN', 'default': False},
		'location': {'type': 'BOOLEAN', 'default': False},
		'fullscreen': {'type': 'BOOLEAN', 'default': True},
		'javascript': {'type': 'BOOLEAN', 'default': True},
		'images': {'type': 'BOOLEAN', 'default': True},
		'adblock': {'type': 'BOOLEAN', 'default': True},
		'mastodon': {'type': 'BOOLEAN', 'default': False},
		'allowhttp': {'type': 'BOOLEAN', 'default': True},
		'capture': {'type': 'BOOLEAN', 'default': False},
		'cookies': {'type': 'BOOLEAN', 'default': False},
		'lastupdate': {'type': 'datetime'}
	},
	'tabs': {
		'id': {'type': 'integer', 'options': ['primary key']},
		'title': {'type': 'text'},
		'url': {'type': 'text'},
		'tabid': {'type': 'integer'}
	},
	'search': {
		'id': {'type': 'integer', 'options': ['primary key']},
		'name': {'type': 'text'},
		'keyword': {'type': 'text', 'options': ['unique']},
		'url': {'type': 'text'}
	}
})


SearchEngines = [
	{'name': 'DuckDuckGo', 'keyword': 'ddg', 'url': 'https://ddg.gg/?q={q}'},
	{'name': 'Google', 'keyword': 'gg', 'url': 'https://google.com/?q={q}'}, #Doesn't actually work right. Remove it?
	{'name': 'Youtube', 'keyword': 'yt', 'url': 'https://www.youtube.com/results?search_query={q}'},
	{'name': 'Weasyl', 'keyword': 'weasyl', 'url': 'https://weasyl.com/search?q={q}'},
	{'name': 'Twitter', 'keyword': 'tw', 'url': 'https://twitter.com/search?q={q}'},
	{'name': 'Github', 'keyword': 'git', 'url': 'https://github.com/search?q={q}'},
	{'name': 'Python Code on Github', 'keyword': 'gitpy', 'url': 'https://github.com/search?q={q}&type=Code&l=Python'},
	{'name': 'Python Documentation', 'keyword': 'py', 'url': 'https://docs.python.org/3/search.html?q={q}'},
	{'name': 'Python Package Index', 'keyword': 'pypi', 'url': 'https://pypi.org/search/?q={q}'},
	{'name': 'PySide2 Documentation', 'keyword': 'pyside', 'url': 'https://doc.qt.io/qtforpython/search.html?q={q}'}
]


def ParseData(table, row):
	tbdata = tables.get(table)
	types = []
	result = []

	if not tbdata:
		logging.error('Invalid table:', table)
		return

	for v in tbdata.values():
		dtype = v.split()[0].upper()

		if dtype == 'BOOLEAN':
			types.append(boolean)

		elif dtype in ['INTEGER', 'INT']:
			types.append(int)

		else:
			types.append(str)

	for idx, v in enumerate(row):
		result.append(types[idx](v) if v else None)

	return row


# Only sqlite for now. Postgresql support planned later
class DB():
	def __init__(self, dbfile=var.database):
		self.db = PooledDB(sqlite3, database=dbfile, maxconnections=50, mincached=5, maxusage=1)
		self.window = None
		self.cache = {}
		self.resclass = {}

		for table in tables.keys():
			self._setup_cache(table)

		# function aliases
		self.get = self.fetch


	def _table_keys(self, table):
		return [key for key in tables[table].keys() if key != 'foreignkey']


	def _setup_cache(self, table):
		self.cache[table] = LRUCache(128)


	def close(self):
		self.db.close()


	def insert(self, table, args):
		placeholders = ",".join(['?' for _ in args.keys()])
		values = tuple(args.values())
		keys = ','.join(args.keys())

		try:
			query = f'INSERT INTO {table} ({keys}) VALUES ({placeholders})'
			self.query(query, values)
			return True
		except Exception as e:
			logging.error(f'Error when inserting: {e}')


	def update(self, table, rowid, data):
		fields = self.GetFields(table, [id])
		newdata = {k: v for k, v in data.items() if k in fields}
		keys = list(newdata.keys())
		values = list(newdata.values())

		if len(newdata) < 1:
				logging.debug('No data provided to update row')
				return False

		query_data = ', '.join(f'{k} = ?' for k in keys)
		query = f'UPDATE {table} SET {query_data} WHERE rowid = {rowid}'

		try:
			self.query(query, values)
			return True
		except Exception as e:
			logging.error(f'Error when updating: {e}')


	def remove(self, table, rowid):
		query = f'DELETE FROM {table} WHERE id = ?'

		try:
			self.query(query, [rowid])
			return True
		except Exception as e:
			logging.error(f'Error when deleting: {e}')


	def fetch(self, table, rowid=None, single=True, sort=None, **kwargs):
		db = self
		table = table if tables.get(table) else table.replace('_migrate', '')
		querysort = f'ORDER BY {sort}'

		with self.GetCursor() as cursor:
			resultOpts = [self, table, cursor]

			if rowid:
				cursor.execute(f"SELECT * FROM {table} WHERE id = ?", [rowid])

			elif kwargs:
				placeholders = [f'{k} = ?' for k in kwargs.keys()]
				values = kwargs.values()

				where = ' and '.join(placeholders)
				query = f"SELECT * FROM {table} WHERE {where} {querysort if sort else ''}"
				cursor.execute(query, list(values))

			else:
				cursor.execute(f'SELECT * FROM {table} {querysort if sort else ""}')

			rows = cursor.fetchall()

			if rows:
				if single:
					return DBResult(rows[0] if single else rows, *resultOpts)

				return [DBResult(row, *resultOpts) for row in rows]

			return None if single else []


	def count(self, table):
		tabledict = tables.get(table)

		if not tabledict:
			logging.debug('Table does not exist:', table)
			return 0

		data = self.query(f'SELECT COUNT(*) FROM {table}')
		return data[0][0]


	@contextmanager
	def GetCursor(self):
		conn = self.db.connection()
		cursor = conn.cursor()

		try:
			yield cursor
		except sqlite3.OperationalError:
			traceback.print_exc()
			conn.rollback()
		finally:
			cursor.close()
			conn.commit()
			conn.close()

	def query(self, string, values=[]):
		with self.GetCursor() as cursor:
			cursor.execute(string, values)
			return cursor.fetchall()


	def GetFields(self, table, ignore_fields=[]):
		keys = []

		for key in tables[table].keys():
			if key not in ignore_fields:
				keys.append(key)

		return keys


	def TableCheck(self, table):
		query = self.query("SELECT count(name) FROM sqlite_master WHERE type=? AND name=?", ['table', table])[0][0]
		return True if query > 0 else False


	def CreateTable(self, table):
		layout = DotDict(tables.get(table))

		if not layout:
			logging.error('Table config doesn\'t exist:', table)
			return

		cmd = f'CREATE TABLE IF NOT EXISTS {table}('
		items = []

		for k, v in layout.items():
			options = ' '.join(v.get('options', []))
			default = v.get('default')
			item = f'{k} {v.type.upper()} {options}'

			if default:
				item += f'DEFAULT {default}'

			items.append(item)

		cmd += ', '.join(items) + ')'

		self.query(cmd)


	def RenameTable(self, table, newname):
		self.query(f'ALTER TABLE {table} RENAME TO {newname}')


	def DropTable(self, table):
		self.query(f'DROP TABLE {table}')


	def UpVersion(self, v):
		row = self.fetch('config', key='version')
		row.value = str(v)
		row.Update()


	def AddColumn(self, table, name, datatype, default=None, options=None):
		query = f'ALTER TABLE {table} ADD COLUMN {name} {datatype.upper()}'

		if default:
			query += f' DEFAULT {default}'

		if options:
			query += f' {options}'

		self.query(query)


	def ParseCookie(self, cookie):
		field = lambda name: getattr(cookie, name)()
		data = {
			'domain': field('domain'),
			'path': field('path'),
			'name': Decode(field('name')),
			'value': Decode(field('value')),
			'expires': field('expirationDate').toTime_t(),
			'http_only': field('isHttpOnly'),
			'secure': field('isSecure'),
			'session': field('isSessionCookie')
		}

		return data


	def CreateCookie(self, row):
		cookie = QNetworkCookie()
		cookie.setDomain(row.domain)
		cookie.setHttpOnly(row.http_only)
		cookie.setName(row.name.encode())
		cookie.setPath(row.path)
		cookie.setSecure(row.secure)
		cookie.setValue(row.value.encode())

		if row.expires > 1:
			date = QDateTime().fromTime_t(row.expires)
			cookie.setExpirationDate(date)

		return cookie


class DBResult(DotDict):
	def __init__(self, row, db, table, cursor):
		super().__init__()

		self.db = db
		self.table = table

		for idx, col in enumerate(cursor.description):
			self[col[0]] = row[idx]


	def __setattr__(self, name, value):
		if name not in ['db', 'table']:
			return self.__setitem__(name, value)

		else:
			return super().__setattr__(name, value)


	def __delattr__(self, name):
		if name not in ['db', 'table']:
			return self.__delitem__(name)

		else:
			return super().__delattr__(name)


	def __getattr__(self, value, default=None):
		options = [value]

		if default:
			options.append(default)

		if value in self.keys():
			val = super().__getitem__(*options)
			return DotDict(val) if isinstance(val, dict) else val

		else:
			return dict.__getattr__(*options)


	# Kept for backwards compatibility. Delete later.
	def asdict(self):
		return self


	def Update(self, data={}):
		self.update(data)
		self.db.update(self.table, self.id, self)


	def Remove(self):
		self.db.remove(self.table, self.id)


def CreateDatabase():
	for name in tables.keys():
		db.CreateTable(name)

	db.insert('config', {'key': 'version', 'value': dbversion})
	db.insert('config', {'key': 'search', 'value': 'ddg'})

	for engine in SearchEngines:
		db.insert('search', engine)


perm = True
backupdb = f'{var.database}.backup-{prev_dbversion}'

if not Path(var.database).exists() and Path(backupdb).exists():
	logging.verbose('Copying backup database')
	shutil.copy(backupdb, var.database)

db = DB(var.database) if perm else DB(':memory:')
