from os.path import join
from pathlib import Path
from urllib.parse import urlparse
from collections import namedtuple

from . import debstable
from .Lib.IzzyLib import logging
from .config import var
from .database import db, put, get, delete
from .functions import GetMD5

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5 import uic


class EditBookmark(QDialog):
	def __init__(self, parent, title=None, url=None, row=None):
		super().__init__(parent)
		self.ui = uic.loadUi(join(var.resources, 'dialog-bookmark.ui'))
		self.parent = parent
		self.setLayout(self.ui.layout())
		self.setWindowTitle('Add/Edit Bookmark')

		self.row = row if row else db.fetch('bookmarks', url=url)

		if not row:
			self.ui.title.setText(title if title else '')
			self.ui.url.setText(url if url else '')
			self.ui.remove.setEnabled(False)

		else:
			self.ui.title.setText(row.name)
			self.ui.url.setText(row.url)
			self.ui.category.setText(row.category)
			self.ui.description.setPlainText(row.description)

		self.ui.save.clicked.connect(lambda : self._button_click('save'))
		self.ui.remove.clicked.connect(lambda : self._button_click('remove'))
		self.ui.discard.clicked.connect(lambda : self._button_click('discard'))


	def _button_click(self, button):
		data = {
			'id': self.row.id if self.row else None,
			'name': self.ui.title.text(),
			'url': self.ui.url.text(),
			'category': self.ui.category.text(),
			'description': self.ui.description.toPlainText()
		}

		actions = {
			'save': {'cmd': put.bookmark, 'args': [data]},
			'remove': {'cmd': db.remove, 'args': ['bookmarks', self.row.id if self.row else None]}
		}

		action = actions.get(button)

		if action:
			action['cmd'](*action.get('args', []), **action.get('kwargs', {}))

		self.close()
		self.deleteLater()
		del self.parent.open_dialogs['bookmarkedit']


class SiteSettings(QDialog):
	def __init__(self, parent, host):
		super().__init__(parent)
		self.ui = uic.loadUi(join(var.resources, 'dialog-sitesettings.ui'))
		self.setLayout(self.ui.layout())
		self.parent = parent
		self.domain = host
		self.setWindowTitle(self.domain.title())

		self.row = get.permissions(self.domain, defaults=True)

		if not self.row.domain:
			self.ui.reset.setEnabled(False)

		else:
			self.ui.adblock.setChecked(self.row.adblock)
			self.ui.fullscreen.setChecked(self.row.fullscreen)
			self.ui.notifications.setChecked(self.row.notification)
			self.ui.cookies.setChecked(self.row.cookies)
			self.ui.capture.setChecked(self.row.capture)
			self.ui.microphone.setChecked(self.row.microphone)
			self.ui.location.setChecked(self.row.location)
			self.ui.camera.setChecked(self.row.camera)
			self.ui.allowhttp.setChecked(self.row.allowhttp)
			self.ui.mastodon.setChecked(self.row.mastodon)

		self.ui.save.clicked.connect(lambda : self._button_click('save'))
		self.ui.reset.clicked.connect(lambda : self._button_click('reset'))
		self.ui.discard.clicked.connect(lambda : self._button_click('discard'))


	def _button_click(self, button):
		data = {
			'id': self.row.id if self.row else None,
			'domain': self.domain,
			'microphone': self.ui.microphone.isChecked(),
			'notification': self.ui.notifications.isChecked(),
			'cookies': self.ui.cookies.isChecked(),
			'capture': self.ui.capture.isChecked(),
			'camera': self.ui.camera.isChecked(),
			'location': self.ui.location.isChecked(),
			'fullscreen': self.ui.fullscreen.isChecked(),
			'adblock': self.ui.adblock.isChecked(),
			'allowhttp': self.ui.allowhttp.isChecked(),
			'mastodon': self.ui.mastodon.isChecked()
		}

		actions = {
			'save': {'cmd': put.permissions, 'args': [self.domain, data]},
			'reset': {'cmd': delete.permission, 'args': [self.domain if self.row else None]}
		}

		action = actions.get(button)

		if action:
			action['cmd'](*action.get('args', []), **action.get('kwargs', {}))

		self.close()
		self.ui.deleteLater()
		del self.parent.open_dialogs['siteedit']


class Toot(QDialog):
	def __init__(self, parent, data=None, reply=None):
		super().__init__(parent)
		self.ui = uic.loadUi(join(var.resources, 'dialog-toot.ui'))
		self.setLayout(self.ui.layout())
		self.setWindowTitle('New Toot')

		self.parent = parent
		self.mastodon = parent.mastodon
		self.accounts = {}
		self.reply = reply
		self.post = None
		self.postid = None

		accts = db.fetch('mastodon', single=False)

		if len(accts) < 1:
			logging.error('No mastodon accounts')
			return

		for row in accts:
			self.accounts[row.fullname] = {
				'displayname': row.displayname,
				'domain': row.domain,
				'apikey': row.apikey,
				'avatar': row.avatar,
				'tootlimit': row.tootlimit
			}

			text = QFontMetrics(self.ui.account.font()).elidedText(row.displayname, Qt.TextElideMode.ElideRight, 100)
			self.ui.account.addItem(QIcon(join(var.profilepath, 'avatars', row.avatar)), text, row.fullname)

			if row.fullname == get.config('defaccount'):
				index = self.ui.account.findData(row.fullname)

				if index < 0:
					logging.error('Could not find item:', row.fullname)
					continue

				self.ui.account.setCurrentIndex(index)

		self.ui.privacy.setItemData(0, 'public')
		self.ui.privacy.setItemData(1, 'unlisted')
		self.ui.privacy.setItemData(2, 'private')
		self.ui.privacy.setItemData(3, 'direct')

		self.ui.post.clicked.connect(lambda *args: self._button_click('post'))
		self.ui.cancel.clicked.connect(lambda *args: self._button_click('cancel'))
		self.ui.warning.textChanged.connect(self._update_text_count)
		self.ui.content.textChanged.connect(self._update_text_count)
		self.ui.account.currentIndexChanged.connect(self._switch_accounts)

		self.textlimit = self.accounts[self.ui.account.currentData()]['tootlimit']

		self._switch_accounts()

		if self.reply:
			self.ui.replytext.setText(self.post.content)
			self.ui.content.document().setPlainText(self.post.acct + ' ' + self.ui.content.toPlainText())
			self.ui.replytext.show()
		else:
			self.ui.replytext.hide()


	def _update_text_count(self):
		text = self.ui.content.toPlainText() + self.ui.warning.text()
		self.ui.count.setText(str(self.textlimit - len(text)))


	def _switch_accounts(self, index=0):
		fullname = self.ui.account.itemData(index)
		self.textlimit = self.accounts[fullname]['tootlimit']
		self._update_text_count()

		if self.reply:
			self._get_post_data()


	def _get_post_data(self):
		currentacct = self.ui.account.currentData()

		data = self.mastodon.GetPostData(self.reply, currentacct)

		if not data:
			return

		newdata = {
			'id': data['id'],
			'username': data['account']['username'],
			'domain': urlparse(data['account']['url']).netloc,
			'content': data['content'],
			'warning': data['spoiler_text'],
			'visibility': data['visibility']
		}

		newdata['acct'] = '@' + newdata['username'] + '@' + newdata['domain']

		self.post = namedtuple('PostData', newdata.keys())(*newdata.values())
		self.postid = self.post.id

		privindex = self.ui.privacy.findData(self.post.visibility)
		warning = self.post.warning if self.post.warning != '' else None
		self.ui.privacy.setCurrentIndex(privindex)

		if warning:
			self.ui.warning.setText('re: ' + warning if not warning.startswith('re:') else warning)

		self.ui.content.setFocus()
		self.ui.setWindowTitle('Reply to ' + self.post.acct if self.postid else 'New Toot')


	def _button_click(self, button):
		warning = self.ui.warning.text()
		content = self.ui.content.toPlainText()

		if button == 'post' and int(self.ui.count.text()) < 0:
			logging.error(f'Over the text limit by {self.ui.count.text()}. Please shorten your message first.')
			return

		data = {
			'warning': warning if warning != '' else None,
			'content': content if content !='' else None,
			'fullname': self.ui.account.currentData(),
			'privacy': self.ui.privacy.currentText().lower()
		}

		if self.postid:
			data['reply'] = self.postid

		actions = {
			'post': {'cmd': self.mastodon.Post, 'kwargs': data},
		}

		action = actions.get(button)

		if action:
			action['cmd'](*action.get('args', []), **action.get('kwargs', {}))

		self.close()
		self.ui.deleteLater()
		del self.parent.open_dialogs['toot']


class NewDownload(QDialog):
	def __init__(self, profile, download):
		super().__init__(profile.window)

		self.ui = uic.loadUi(join(var.resources, 'dialog-download.ui'))
		self.setWindowTitle('Download File: ')
		self.setModal(True)
		self.setFixedHeight(200)
		self.setLayout(self.ui.layout())

		self.download = download
		self.md5 = GetMD5(self.download.item.url().toString() + '.md5') if get.config('automd5') else None

		download.md5 = self.md5

		self.profile = profile
		self.finished = False

		if debstable:
			## This is a half-assed fix and needs to be improved later
			self.file = Path(self.download.item.url().path()).name
		else:
			self.file = self.download.item.suggestedFileName()

		self.dir = get.config('dlpath', var.downloadpath)
		self.ui.md5.setText(self.md5)

		self.ui.filename.returnPressed.connect(lambda: self._handle_close_button('save'))
		self.ui.filename.textChanged.connect(self._filename_changed)
		self.ui.md5.editingFinished.connect(self._hash_changed)
		self.ui.pickname.clicked.connect(self._pick_filename)
		self.ui.reset.clicked.connect(self._reset)
		self.ui.cancel.clicked.connect(lambda *args: self._handle_close_button('cancel'))
		self.ui.save.clicked.connect(lambda *args: self._handle_close_button('save'))

		self._reset()
		self.show()

		url = self.download.item.url().toString()
		urltext = QFontMetrics(self.ui.urlfilename.font()).elidedText(url, Qt.TextElideMode.ElideMiddle, self.ui.urlfilename.width() - 20)
		self.ui.urlfilename.setText(urltext)


	def _filename_changed(self, filename):
		if Path(filename).exists():
			self.ui.save.setText('Overwrite')

		else:
			self.ui.save.setText('Save')


	def _hash_changed(self):
		text = self.ui.md5.text()
		self.download.item.md5 = None if text == '' else text


	def _pick_filename(self):
		name, _ = QFileDialog.getSaveFileName(caption='Select Download Filename', directory=join(self.dir, self.file), filter=f'*{Path(self.file).suffix}')

		if name != '':
			self.ui.filename.setText(str(filename))


	def _reset(self):
		filepath = join(self.dir, self.file)
		self.ui.filename.setText(filepath)


	def _handle_close_button(self, button):
		url = self.download.item.url().toString()

		if button == 'cancel':
			logging.verbose('Cancelled download of', url)
			self.download.item.cancel()

		elif button == 'save':
			logging.verbose('Starting download of', url)
			self.SetFilename()
			self.download.item.accept()
			self.download.item.finished.connect(lambda: self.profile._download_finished(self.download))

		self.finished = True
		self.close()
		self.deleteLater()


	def SetFilename(self):
		filename = Path(join(self.ui.filename.text()))

		if debstable:
			self.download.item.setPath(str(filename))

		else:
			self.download.item.setDownloadDirectory(str(filename.parent))
			self.download.item.setDownloadFileName(str(filename.name))


class HttpAuth(QDialog):
	def __init__(self, url, authenticator, parent):
		super().__init__(parent.window)

		self.ui = uic.loadUi(join(var.resources, 'dialog-authentication.ui'))
		self.ui.savepass.setEnabled(False)
		self.setLayout(self.ui.layout())

		self.webview = parent
		self.url = url
		self.auth = authenticator
		self.username = None
		self.password = None
		self.result = False

		text = self.ui.realm.text().format(host=url.host(), msg=self.auth.realm())
		self.ui.realm.setText(text)

		self.ui.login.clicked.connect(lambda : self._handle_close_button('login'))
		self.ui.cancel.clicked.connect(self._handle_close_button)


	def _handle_close_button(self, button=None):
		if button == 'login':
			self.auth.setUser(self.ui.username.text())
			self.auth.setPassword(self.ui.password.text())
			self.result = True

		self.close()
		self.deleteLater()
