import os, sys, json, platform, struct, hashlib, traceback

from os.path import isfile, join, dirname
from os import environ, getcwd
from xml.sax.saxutils import unescape
from urllib.parse import urlparse
from datetime import datetime
from configparser import ConfigParser
from pathlib import Path
from collections import OrderedDict

import urllib3

from plyer import notification
from publicsuffixlist import PublicSuffixList
from .Lib.IzzyLib import logging

from . import __version__, __application__
from .config import var
from .database import put

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtWebEngineWidgets import *


class Notifications(object):
	def __init__(self):
		pass

	def New(self, title, message, icon=None):
		options = {
			'app_name': 'QtWeb',
			'title': title,
			'message': message
		}

		if icon:
			options['app_icon'] = icon

		notification.notify(**options)

	def Info(self, message):
		self.New('Info', message)

	def Error(self, message):
		self.New('Error', message)


def Icon(name, size=None):
	filename = join(var.iconpath, name + '.svg')

	if not isfile(filename):
		logging.error('Invalid icon name:', name)
		icon = QIcon(join(var.iconpath, 'missing.svg'))

	icon = QIcon(filename)

	if size:
		icon = QIcon(icon.pixmap(16))

	return icon


def AddShortcut(key, action, widget):
	focusurl = QShortcut(QKeySequence(key), widget)
	focusurl.activated.connect(action)
	focusurl.setEnabled(True)


def Fetch(resource, out='json'):
	resp = None

	try:
		resp = httpclient.request('GET', resource)
	except urllib3.exceptions.MaxRetryError:
		logging.verbose(f'Failed to connect to {urlparse(resource).netloc}')
	except Exception as e:
		traceback.print_exc()
		logging.error('Connection error:', e)
		return

	if not resp:
		logging.verbose(f'Failed to fetch {resource}')
		return

	if resp.status not in [200, 202]:
		logging.verbose(f'Failed to fetch {resource}:', resp.status)
		return

	try:
		if out == 'json':
			data =  json.loads(resp.data.decode())
			return data

		elif out == 'text':
			return resp.data.decode()

		else:
			return resp.data

	except Exception as e:
		logging.verbose(e)
		return


def FetchNodeinfo(domain):
	nodeinfo = Fetch(f'https://{domain}/nodeinfo/2.0.json')

	if not nodeinfo or (isinstance(nodeinfo, dict) and nodeinfo.get('error')):
		logging.verbose(f'Wrong nodeinfo url for {domain}. Finding correct one...')
		wk_url = f'https://{domain}/.well-known/nodeinfo'
		well_known = Fetch(wk_url)

		if not well_known:
			return {}

		ni_url = None

		try:
			for link in well_known['links']:
				if link['rel'] == 'http://nodeinfo.diaspora.software/ns/schema/2.0':
					ni_url = link['href']

		except Exception as e:
			logging.verbose(f'Invalid nodeinfo data for {domain}. Error: {e}')
			return

		if not ni_url:
			logging.verbose(f'failed to find nodeinfo url for {domain}')
			return

		nodeinfo = Fetch(ni_url)

		if not nodeinfo:
			logging.verbose(f'Failed to fetch nodeinfo for {domain}')
			nodeinfo = {}

	return nodeinfo


def NewDesktopEntry(path):
	filename = Path(path) if type(path) != Path else path
	deskfilename = 'xyz.barkshark.qtweb.desktop'

	mimetypes = [
		'text/html',
		'text/xml',
		'application/xhtml+xml',
		'application/xml',
		'x-scheme-handler/http',
		'x-scheme-handler/https'
	]

	deskfile = ConfigParser(interpolation=None)
	deskfile.optionxform = str
	deskfile['Desktop Entry'] = {
		'Name': 'QtWeb Dev',
		'Comment': 'Browse the World Wide Web with less bullshit',
		'GenericName': 'WebBrowser',
		'Exec': '/usr/bin/python3 -m qtweb %U',
		'Icon': join(var.iconpath, 'icon.svg'),
		'Type': 'Application',
		'Categories': 'Network;WebBrowser;',
		'MimeTypes': ''.join(mime + ';' for mime in mimetypes),
		'StartupWMClass': 'QtWeb',
		'StartupNotify': 'true'
	}

	if getcwd() == dirname(var.scriptpath):
		print('heck')
		deskfile['Desktop Entry']['Path'] = dirname(var.scriptpath)

	with filename.joinpath(deskfilename).open('w') as fd:
		deskfile.write(fd)

	return deskfilename


def GetSoftware(domain):
	if not domain or domain == '':
		return

	ni = FetchNodeinfo(domain)
	ismastodon = ni.get('software', {}).get('name') == 'mastodon'
	
	put.permissions(domain, {'mastodon': ismastodon})


def GetMD5(url):
	if not url:
		return

	data = Fetch(url, out='text')

	if not data:
		return

	return data.split()[0]


def MD5Check(filename, hashdigest, blocksize=65536):
	'only works with one hash. will fix later'
	hash = hashlib.md5()
	with filename.open("rb") as f:
		for block in iter(lambda: f.read(blocksize), b""):
			hash.update(block)
	return hash.hexdigest() == hashdigest


def UserAgentGen():
	#bit = '64' if platform.machine().endswith('64') else '32'
	os = platform.system().lower()

	bit = 'x86_64' if struct.calcsize("P") == 8 else 'i686'
	base = 'AppleWebKit/537.36 (KHTML, like Gecko) Version/13.0 Safari/537.36'

	osbits = {
		'windows': {
			'x86_64': 'WOW64',
			'i686': 'Win64'
		}
	}

	extra = ''

	if os == 'linux':
		if environ.get('WAYLAND_DISPLAY'):
			extra = 'wayland; '
		elif environ.get('DISPLAY'):
			extra = 'X11; '

	osbit_opts = osbits.get(os)

	if not osbit_opts:
		osbit = bit

	else:
		osbit = osbit_opts.get(bit)

	if os == 'windows':
		os == 'Windows NT 10.0'

	return f'Mozilla/5.0 ({extra}{os.title()}; {osbit}) {base} {__application__}/{__version__}'


def Decode(byte):
	return TextCodec.toUnicode(byte)


## subclass QWebEngineScript to add extra functions
class WebEngineScript(QWebEngineScript):
	def __init__(self, collection=None, filename=None, user=True):
		super().__init__()

		self.collection = collection
		self.user = user
		self.filename = None
		self.enabled = False

		if filename:
			self.initLoad(filename)


	def initLoad(self, filename):
		self.filename = Path(filename)

		with self.filename.open() as fd:
			self.setSourceCode(fd.read())


	def enable(self):
		if not self.enabled:
			logging.verbose('Enabling script:', self.name())
			self.enabled = True
			return self.collection.insert(self)


	def disable(self):
		if self.enabled:
			logging.debug('Disabling script:', self.name())
			self.enabled = False
			return self.collection.remove(self)


## subclass QMenu because there's no easy way to get specific items
class Menu(QMenu):
	def __init__(self, parent, title=None):
		super().__init__(parent)
		self.items = OrderedDict()

		if title:
			self.setTitle(title)


	def AddItem(self, label, callback, *args, icon=None, **kwargs):
		if self.items.get(label):
			logging.error('An action with this name already exists:', label)
			return

		self.items[label] = (self.addAction(label, lambda : callback(*args, **kwargs)))

		if icon:
			self.items[label].setIcon(icon if type(icon) == QIcon else QIcon(icon))

		return True


	def AddItemAfter(self, indexLabel, label, callback, *args, **kwargs):
		index = self.items.get(indexLabel)
		newdict = self.items.__class__()

		if not index:
			self.error('Not a menu action:', indexLabel)
			return

		self.clear()

		for name, action in self.items.items():
			newdict[name] = action

			if name == indexLabel:
				newdict[label] = self.AddItem(label, callback, *args, **kwargs)

		self.items.clear()
		self.items.update(newdict)
		return True


	def AddActionAfter(self, action, indexAction):
		label = action.text()
		indexLabel = index.text()
		index = self.items.get(indexLabel)
		newdict = self.items.__class__()

		if not index:
			self.error('Not a menu action:', indexLabel)
			return

		self.clear()

		for name, action in self.items.items():
			newdict[name] = self.addAction(name, action)

			if name == indexLabel:
				newdict[label] = self.AddItem(label, callback, *args, **kwargs)

		self.items.clear()
		self.items.update(newdict)
		return True


	def MoveItem(self, label, indexLabel):
		action = self.items.get(label)
		index = self.items.get(indexLabel)
		newdict = self.items.__class__()

		if None in [action, index]:
			logging.error(f'Invalid action or index action: {label}: {action}, {indexLabel}: {index}')
			return

		self.clear()

		for name, action in self.items.items():
			newdict[name] = self.addAction(name, action)

			if name == indexLabel:
				newdict[label] = self.addAction(label, action)

		self.items.clear()
		self.items.update(newdict)
		return True


	def RemoveItem(self, label):
		action = self.items.get(label)

		if not action:
			logging.error('Invalid action:', label)
			return

		self.removeAction(action)
		action.DeleteLater()
		del action
		return True


	def AddSeparator(self, name='separator'):
		item = self.items.get(name)
		count = 0

		while item:
			count += 1
			item = self.items.get(name + str(count))

		label = name + str(count) if count > 0 else name
		self.items[label] = self.addSeparator()
		return True


	def MoreThan(self, num):
		return len(self.items) > num


	def LessThan(self, num):
		return len(self.items) < num


def DateTime(timestamp):
	if timestamp in ['0', 0, False]:
		return 0

	if type(timestamp) == int:
		return datetime.fromtimestamp(str(timestamp))

	elif type(timestamp) == str:
		return datetime.fromisoformat(timestamp)

	elif type(timestamp) == datetime:
		return timestamp

	else:
		logging.error('Wrong type for datetime object:', type(timestamp))


def PrintMethods(obj):
	for line in dir(obj):
		if not line.startswith('_'):
			print(line)


def ParseSoundcloudInfo(data):
	if not data:
		return False

	data = data.split(':::')

	if len(data) != 6:
		logging.debug('Wrong number of soundcloud data slices:', data)
		return False

	artist, artistLink, title, titleLink, progress, length = data

	info = DotDict({
		'artist': unescape(artist),
		'artistLink': artistLink,
		'title': unescape(title.replace('Current track: ', '')),
		'titleLink': titleLink,
		'progress': progress,
		'length': length
	})

	return info


class DotDict(dict):
	def __init__(self, value=None, **kwargs):
		super().__init__()
		self.__setattr__ = dict.__setitem__
		self.__delattr__ = dict.__delitem__
		self.__getitem__ = self.__getattr__

		if value.__class__ == str:
			self.FromJson(value)

		elif value.__class__ in [dict, DotDict]:
			self.update(value)

		elif value:
			raise TypeError('The value must be a JSON string, dict, or another DotDict object, not', value.__class__)

		if kwargs:
			print('From kwargs')
			self.update(kwargs)


	def ToJson(self, **kwargs):
		return self.__str__(**kwargs)


	def FromJson(self, string):
		data = json.loads(string)
		self.update(data)


	def __parse_item(self, data):
		return DotDict(data) if type(data) == dict else data


	def items(self):
		data = []

		for k, v in super(DotDict, self).items():
			value = self.__parse_item(v)
			data.append((k, value))

		return data


	def __str__(self, **kwargs):
		return json.dumps(self, **kwargs)


	def __getattr__(self, value, default=None):
		val = self.get(value, default) if default else self[value]
		return self.__parse_item(val)


def ExceptionHandler(exctype, value, tb):
	sys._excepthook(exctype, value, tb)

	# Apparently the line above is enough to not bring down the whole browser
	# Gonna keep custom handler here for now though
	#lines = [line for line in traceback.format_exception(exctype, value, tb)]

	#if lines[-1] in ['\n', '\n ']:
		#lines.remove(lines)

	#error = ''.join(lines)
	#print(f'\n------------\n{error}------------')


TextCodec = QTextCodec.codecForMib(106)
useragent = UserAgentGen()
httptimeout = urllib3.Timeout(connect=5, read=2)
httpclient = urllib3.PoolManager(num_pools=100, timeout=httptimeout, headers={'user-agent': f'{__application__}/{__version__}'})
psl = PublicSuffixList()
