import sys, os, json, signal, asyncio, tracemalloc

from urllib.parse import urlparse, quote_plus
from os.path import join, abspath
from pathlib import Path

from .Lib.IzzyLib import logging
from .Lib.IzzyLib.misc import boolean

from . import isWindows, debstable, dialogs, tabs, __version__
from .urlhandlers import LocalHandler, RegisterScheme
from .webview import WebEngineView
from .webprofile import WebEngineProfile
from .config import var
from .database import db, get, put
from .mastodon import Mastodon
from .mbus import Client, Server
#from .themes import ColorTheme
from .arguments import args
from .functions import (
	Icon, AddShortcut, UserAgentGen, DotDict,
	ExceptionHandler, Notifications, Menu
)

if isWindows:
	import ctypes

	OpenProcess = ctypes.windll.kernel32.OpenProcess
	CloseHandle = ctypes.windll.kernel32.CloseHandle
	SetAppId = ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID

	## Set the application icon on windows
	SetAppId('xyz.barkshark.qtweb'+__version__)

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtWebEngineWidgets import *
from PyQt5 import uic


## Prevent errors from bringing down the whole browser
sys._excepthook = sys.excepthook
sys.excepthook = ExceptionHandler

logging.setConfig({'level': get.config('loglevel', 'INFO')})
Window = None


class Browser(QMainWindow):
	def __init__(self, app, urls=[]):
		super().__init__()
		global Window
		Window = self

		if get.config('tracestart'):
			logging.verbose('Starting tracemalloc')
			tracemalloc.start()
			self.snapshot = tracemalloc.take_snapshot()

		else:
			self.snapshot = None

		if not debstable:
			RegisterScheme(var.local)

		db.window = self

		self.app = app
		self.default_style = app.style().objectName().lower()
		self.default_palette = app.style().standardPalette()
		#self.themes = ColorTheme(self)
		self.clipboard = app.clipboard()
		self.threadpool = QThreadPool()
		self.notif = Notifications()
		self.statedata = {
			'width': 800,
			'height': 600,
			'x': 100,
			'y': 100,
			'maximized': True
		}

		self.statedata.update(json.loads(get.config('state', "{}")))

		## Setup empty variables for use later
		self.snapshot = None
		self.server = None
		self.socket = None
		self.urlhover = None
		self.fullscreen = False
		self.shutdown = False
		self.widgettabs = DotDict()
		self.open_dialogs = {}

		self.profile = WebEngineProfile(self)
		self.local_scheme_handler = LocalHandler(self)
		self.mastodon = Mastodon()

		#logging.setConfig({'systemnotif': self.notif})

		#app.setStyle('fusion')
		#self.themes.ChangeTheme()

		# Re-enable when color themes are fixed
		#if theme != 'Default':
			#self.themes.ChangeTheme(theme)

		self.ui = uic.loadUi(join(var.resources, 'window.ui'))
		self.setCentralWidget(self.ui)
		self.setWindowTitle('QtWeb')

		self.ui.url.installEventFilter(UrlBarEsc(self))
		self.ui.menu.setMenu(self._main_menu())

		self.tabbar = TabBar(self)
		self.ui.tabs.setTabBar(self.tabbar)
		self.ui.tabs.tabCloseRequested.connect(self.CloseTab)
		self.ui.tabs.currentChanged.connect(self._change_tab)

		## Navigation bar icons
		self.ui.back.setIcon(Icon('back'))
		self.ui.next.setIcon(Icon('next'))
		self.ui.stop.setIcon(Icon('stop'))
		self.ui.refresh.setIcon(Icon('refresh'))
		self.ui.home.setIcon(Icon('home'))
		self.ui.go.setIcon(Icon('go'))
		self.ui.newtab.setIcon(Icon('newtab'))
		self.ui.menu.setIcon(Icon('menu'))

		## Status bar icons
		self.ui.siteedit.setIcon(Icon('permissions'))
		self.ui.bookmarkedit.setIcon(Icon('favourite'))
		self.ui.toot.setIcon(Icon('toot'))

		## Connect widgets to their actions
		self.ui.back.clicked.connect(lambda : self._button_click('back'))
		self.ui.next.clicked.connect(lambda : self._button_click('next'))
		self.ui.stop.clicked.connect(lambda : self._button_click('stop'))
		self.ui.refresh.clicked.connect(lambda : self._button_click('refresh'))
		self.ui.home.clicked.connect(lambda : self._button_click('home'))
		self.ui.url.returnPressed.connect(lambda : self._button_click('go'))
		self.ui.url.textEdited.connect(lambda text: setattr(self.ui.tabs.currentWidget(), 'currenttext', text))
		self.ui.url.editingFinished.connect(lambda *args: self.ui.url.setCursorPosition(0))
		self.ui.go.clicked.connect(lambda : self._button_click('go'))
		self.ui.newtab.clicked.connect(lambda : self._button_click('newtab'))

		self.ui.bookmarkedit.clicked.connect(lambda : self._open_dialog('bookmarkedit'))
		self.ui.siteedit.clicked.connect(lambda : self._open_dialog('siteedit'))
		self.ui.toot.clicked.connect(lambda : self._open_dialog('toot'))

		## Keyboard shortcuts
		AddShortcut('Ctrl+L', self._focus_urlbar, self)
		AddShortcut('Ctrl+T', lambda *args: self.NewWebTab(switch=True), self)
		AddShortcut('Ctrl+R', lambda *args: self.ui.tabs.currentWidget().reload(), self)
		AddShortcut('Ctrl+W', lambda *args: self.CloseTab(), self)
		AddShortcut('Ctrl+N', lambda *args: self.NewWebTab('https://www.youtube.com/watch?v=dQw4w9WgXcQ', switch=True), self)
		AddShortcut('Ctrl+F', lambda *args: self.ui.tabs.currentWidget().searchbar.Search('toggle'), self)
		AddShortcut('F3', lambda *args: self.ui.tabs.currentWidget().searchbar.Search('toggle'), self)
		AddShortcut('F5', lambda *args: self.ui.tabs.currentWidget().reload(), self)
		AddShortcut('F11', lambda *args: self.Fullscreen(), self)

		for url in urls:
			self.NewWebTab(url)

		for row in db.fetch('tabs', sort='tabid', single=False):
			self.NewWebTab(row.url)

		activetab = int(get.config('activetab', 0))
		self.ui.tabs.setCurrentIndex(activetab)
		self._prevent_empty_tabview()

		## Restore the window position and size
		self.setGeometry(
			self.statedata['x'],
			self.statedata['y'],
			self.statedata['width'],
			self.statedata['height']
		)

		if self.statedata['maximized']:
			self.showMaximized()


	def _main_menu(self):
		menu = QMenu('QtWeb Menu', self.ui.menu)
		menu.addAction('Find In Page', lambda : self.ui.tabs.currentWidget().Search('toggle'))
		menu.addAction('Fullscreen', self.Fullscreen)
		menu.addSeparator()
		menu.addAction('Bookmarks', lambda *args: self.NewWidgetTab('bookmarks'))

		# I should've disabled this a long time ago
		#menu.addAction('History', lambda *args: self.NewWidgetTab('history'))

		menu.addAction('Preferences', lambda *args: self.NewWidgetTab('preferences'))

		return menu


	def _focus_urlbar(self, *args):
		self.ui.url.setFocus()
		self.ui.url.selectAll()


	def _button_click(self, button):
		widget = self.ui.tabs.currentWidget()
		homepage = get.config('homepage', 'https://ddg.gg')

		if button == 'newtab':
			self.NewWebTab(switch=True)
			return

		if hasattr(widget, 'LoadUrl'):
			actions = {
				'back': {'cmd': widget.back},
				'next': {'cmd': widget.forward},
				'stop': {'cmd': widget.stop},
				'refresh': {'cmd': widget.reload},
				'home': {'cmd': widget.LoadUrl, 'args': [homepage]},
				'go': {'cmd': widget.LoadUrl, 'args': [self.ui.url.text()]}
			}

		#else:
			#actions = {'newtab': {'cmd': self.NewWebTab, 'kwargs': {'switch': True}}}

		action = actions.get(button)

		if button == 'stop':
			self._update_webaction_buttons('reload', True)
			self._update_webaction_buttons('stop', False)

		if action:
			action['cmd'](*action.get('args', []), **action.get('kwargs', {}))


	def _open_dialog(self, dialog):
		if dialog == 'toot':
			if len(db.fetch('mastodon', single=False)) < 1:
				self.notif.Error('No mastodon accounts')
				logging.error('No mastodon accounts')
				return

		widget = self.open_dialogs.get(dialog)

		if widget:
			widget.activateWindow()
			return

		webview = self.ui.tabs.currentWidget()
		actions = {
			'bookmarkedit': {
				'cmd': dialogs.EditBookmark,
				'kwargs': {
					'title': webview.title(), 'url': webview.url().toString(), 'row': db.get('bookmarks', url=webview.url().toString())
				}
			},
			'siteedit': {
				'cmd': dialogs.SiteSettings,
				'args': [webview.url().host()]
			},
			'toot': {
				'cmd': dialogs.Toot,
				'args': [self]
			}
		}

		action = actions.get(dialog)

		if not action:
			return

		self.open_dialogs[dialog] = action['cmd'](self, *action.get('args', []), **action.get('kwargs', {}))
		self.open_dialogs[dialog].show()


	def _update_webaction_buttons(self, name=None, state=None):
		webview = self.ui.tabs.currentWidget()
		actions = {
			'back': self.ui.back,
			'forward': self.ui.next,
			'reload': self.ui.refresh,
			'stop': self.ui.stop
		}

		main_widgets = [
			self.ui.home,
			self.ui.url,
			self.ui.go,
			self.ui.siteedit,
			self.ui.bookmarkedit
		]

		if webview.windowTitle():
			widgets = list(actions.values()) + main_widgets
			for widget in widgets:
				widget.setEnabled(False)

		elif not name:
			for widget in main_widgets:
				widget.setEnabled(True)

			for name, widget in actions.items():
				if name in webview.webaction_state.keys():
					widget.setEnabled(webview.webaction_state[name])

		else:
			action = actions.get(name)
			currstate = webview.webaction_state[name] if state == None else state

			if not action:
				return

			action.setEnabled(currstate)


	def _change_tab(self, index):
		widget = self.ui.tabs.widget(index)

		if not widget:
			return

		url = widget.currenttext if not widget.windowTitle() else widget.windowTitle()
		self.ui.url.setText(url)
		self.ui.url.setCursorPosition(0)
		self._update_webaction_buttons()


	def _prevent_empty_tabview(self):
		if self.ui.tabs.count() < 1:
			self.NewWebTab()


	def closeEvent(self, event=None):
		## Make sure this function only fires once
		if self.shutdown:
			return

		self.shutdown = True

		if event:
			event.accept()

		lock = Path(var.lock)
		opentabs = []
		tabrange = list(reversed(range(0, self.ui.tabs.count())))
		put.config('activetab', str(self.ui.tabs.currentIndex()))
		self.SaveWinState()

		for tabid in ([0] if len(tabrange) < 1 else tabrange):
			widget = self.ui.tabs.widget(tabid)

			if widget.windowTitle():
				continue

			title = widget.title()
			url = widget.url().toString()

			if widget.devpage.view:
				widget.devpage.closeEvent()
				widget.devpage.destroy()

			opentabs.append(url)
			put.tab(tabid, url, title)

			self.CloseTab(tabid, shutdown=True)

		for row in db.fetch('tabs', single=False):
			if row.url not in opentabs:
				db.remove('tabs', row.id)

		db.close()
		logging.verbose('Saved state and tabs. Exiting...')

		for srv in [self.server, self.socket]:
			if srv:
				srv.stop()

		if lock.exists():
			lock.unlink()


	def SaveWinState(self, width=None, height=None, x=None, y=None):
		winstate = self.windowState() in [Qt.WindowActive, Qt.WindowNoState]
		maximized = self.windowState() == Qt.WindowMaximized
		fullscreen = self.FullscreenState()
		maxnew = maximized != self.statedata['maximized']

		if winstate:
			self.statedata['x'] = self.geometry().x() if not x else x
			self.statedata['y'] = self.geometry().y() if not y else y
			self.statedata['width'] = self.width() if not width else width
			self.statedata['height'] = self.height() if not height else height

		if maxnew and not fullscreen:
			self.statedata['maximized'] = maximized

		if maxnew or winstate:
			statejson = json.dumps(self.statedata)

			if statejson != get.config('state'):
				logging.verbose('Saving window state data')
				put.config('state', statejson)


	def FullscreenState(self):
		return self.windowState() == Qt.WindowFullScreen


	def Fullscreen(self, action=None):
		maximized = self.statedata['maximized']
		## If no action specified, toggle
		if action == None:
			action = not self.FullscreenState()

		if action:
			self.showFullScreen()
			#self.ui.navbar.hide()
			#self.tabbar.hide()
			#self.ui.statusbar.hide()

		elif not action:
			self.ui.statusbar.show()
			self.tabbar.show()
			self.ui.navbar.show()

			if maximized:
				self.showNormal()
				self.showMaximized()

			else:
				self.showNormal()


	def NewWebTab(self, url=None, switch=False, nexttab=False):
		webview = WebEngineView(self, url if url else get.config('homepage', 'https://ddg.gg'))

		if nexttab and get.config('nexttab', True):
			current_index = self.ui.tabs.currentIndex() + 1
			index = self.ui.tabs.insertTab(current_index, webview, 'New Tab')

		else:
			index = self.ui.tabs.addTab(webview, 'New Tab')

		if switch:
			self.ui.tabs.setCurrentIndex(index)

			if not url:
				self.ui.url.setFocus()
				self.ui.url.selectAll()

		return webview


	def NewWidgetTab(self, widgetname, switch=True):
		widgetname = widgetname.title()
		currwidget = self.widgettabs.get(widgetname)

		if currwidget:
			index = self.ui.tabs.indexOf(currwidget)
			self.ui.tabs.setCurrentIndex(index)
			return

		widgettabs = {
			'Bookmarks': tabs.Bookmarks,
			'History': tabs.History,
			'Downloads': tabs.Downloads,
			'Preferences': tabs.Preferences
		}

		widget = widgettabs.get(widgetname)

		if not widget:
			logging.error('Failed to open widget tab:', widgetname)
			return

		self.widgettabs[widgetname] = widget(self)
		widget = self.widgettabs[widgetname]
		widget.setWindowTitle(widgetname)
		index = self.ui.tabs.addTab(widget, widget.windowTitle())

		if switch:
			self.ui.tabs.setCurrentIndex(index)

		return widget


	def CloseTab(self, index=None, shutdown=False):
		widgettabs = self.widgettabs
		tabs = self.ui.tabs

		if index == None:
			index = -1

		index = int(index)

		if index < 0:
			index = tabs.currentIndex()

		widget = tabs.widget(index)
		tabtitle = widget.windowTitle()

		if widgettabs.get(tabtitle):
			del widgettabs[tabtitle]

		tabs.removeTab(index)
		widget.deleteLater()

		if not shutdown:
			self._prevent_empty_tabview()
			tabs.currentWidget().setFocus()


class UrlBarEsc(QObject):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.window = args[0]


	def eventFilter(self, obj, event):
		if event.type() == QEvent.KeyPress:
			if event.key() == 16777216:
				tab = self.window.ui.tabs.currentWidget()
				url = tab.url().toString()
				tab.currenttext = url
				tab.setFocus()

				self.window.ui.url.setText(url)

		return QObject.eventFilter(self, obj, event)


## Custom tabbar class
class TabBar(QTabBar):
	def __init__(self, window):
		super().__init__(window)
		self.setElideMode(Qt.TextElideMode.ElideRight)
		self.setAutoHide(int(get.config('hidetabs', 1)))
		self.setMovable(True)
		self.setTabsClosable(True)
		self.setUsesScrollButtons(True)
		self.setSelectionBehaviorOnRemove(QTabBar.SelectPreviousTab)
		self.setDocumentMode(True)
		self.window = window

		self.setStyleSheet('''
			QTabBar::tab:selected {font: bold;}
			QTabBar::tab-bar {alignment: center;}
		''')

	## Set tabs to even take up full width
	def tabSizeHint(self, index):
		compensation = 0 #I'll need this later
		width = (self.parent().width() - compensation) / self.count()
		height = QTabBar.tabSizeHint(self, index).height()
		return QSize(width, height)


def GetPid(path):
	if not path.exists():
		return False

	result = None

	with path.open() as fd:
		data = fd.read()

		if data == '':
			result = False

		pid = int(data)

		if isWindows and result == None:
			handle = OpenProcess(1, False, pid)
			
			if handle == 0:
				result = False
			
			else:
				result = True
			
			CloseHandle(handle)

		elif result == None:
			try:
				os.kill(pid, 0)
			except ProcessLookupError:
				result = False
			except PermissionError:
				result = True
			else:
				result = True

	if result == False:
		path.unlink()

	return result


def SetPid(path):
	with path.open('w') as fd:
		fd.write(str(os.getpid()))


def main():
	lock = Path(var.lock)
	app = QApplication(sys.argv)

	app.setApplicationName('QtWeb')
	app.setOrganizationName('Barkshark')
	app.setOrganizationDomain('barkshark.xyz')
	app.setWindowIcon(Icon('icon'))

	if get.config('remotedebug', False):
		debugport = get.config('remotedebugport', '8018')
		debughost = 'still gotta sort this out'
		os.environ['QTWEBENGINE_REMOTE_DEBUGGING'] = f'127.0.0.1:{remotedebugport}'

	host = get.config('mhost', 'localhost')
	port = get.config('mport', 8008)

	if isWindows and host != False:
		if host == '0.0.0.0':
			host = 'localhost'
		mbus = Client(host, port)

	else:
		mbus = Client(sock=Path(var.profilepath, 'mbus.sock'))

	if mbus.setup():
		logging.info('QtWeb already open. Bringing up window...')

		for url in args.urls:
			mbus.NewTab(url, True if url == args.urls[-1] else False)

		mbus.Present()
		app.quit()
		return

	if GetPid(lock):
		logging.error('QtWeb is already running with this profile:', var.profile)
		app.exit(1)
		return

	SetPid(lock)


	window = Browser(app)
	window.server = Server(window)
	app.aboutToQuit.connect(window.close)

	if not isWindows:
		window.socket = Server(window, True)
		window.socket.start()

	for url in args.urls:
		window.NewWebTab(url, True if url == args.urls[-1] else False)

	if get.config('mhost', '0.0.0.0'):
		window.server.start()

	## Run SaveWinState every second. It doesn't do anything if the state hasn't changed
	## This was originally here to allow python to handle signals, so why not take advantage of it?
	timer = QTimer()
	timer.start(1000)
	timer.timeout.connect(window.SaveWinState)
	signal.signal(signal.SIGTERM, lambda *x: app.quit())
	signal.signal(signal.SIGINT, lambda *x: app.quit())

	window.show()

	sys.exit(app.exec())
