import json, shutil

from urllib.parse import urlparse
from os.path import join, isfile, basename
from os import makedirs, remove

from .Lib.IzzyLib import logging
from .config import var
from .database import db, put, get
from .functions import httpclient

from PyQt5.QtCore import *
from PyQt5.QtGui import *


# I was thinking about replacing this with Mastodon.py, but it works fine, so I'm not sure now
class Mastodon(object):
	def __init__(self):
		self.accounts = {}


	def _request(self, api, fullname=None, method='GET', data=None, domain=None, apikey=None):
		row = db.fetch('mastodon', fullname=fullname) if fullname else {}

		if not row:
			row = {}

		key = row.get('apikey') if not apikey else apikey
		domain = row.get('domain') if not domain else domain

		if not domain:
			logging.error('Domain not provided for mastodon api')
			return

		if method.upper() not in ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']:
			logging.error(f'Invalid method for mastodon api:', method)
			return

		options = {
			'headers': {
				'Content-Type': 'application/json'
			}
		}

		if key:
			options['headers']['Authorization'] = f'Bearer {key}'

		if data:
			options['body'] = json.dumps(data).encode('utf-8')

		resp = httpclient.request(method.upper(), f'https://{domain}/{api}', **options)
		data = resp.data

		if resp.status not in [200, 202]:
			logging.error('Failed to contact mastodon:', resp.status, resp.data)
			return False

		try:
			return json.loads(data)
		except Exception as e:
			logging.debug(str(e))
			return False


	def GetAvatar(self, url, current=None):
		avipath = join(var.profilepath, 'avatars')
		makedirs(avipath, exist_ok=True)
		parsed = urlparse(url)
		filename = basename(parsed.path)
		filepath = join(avipath, filename)
		oldfilepath = join(avipath, current) if current else None

		if current and filename == current and isfile(oldfilepath):
			logging.debug(f'{filename} already exists. Not updating.')
			return current

		if not isfile(filepath):
			try:
				response = httpclient.request('GET', url)

				with open(filepath, 'wb') as fd:
					data = response.data

					if response.status in [200, 202]:
						fd.write(data)

					else:
						logging.error(f'Error when grabbing avatar: {response.status} {response.data}')
						return current

			except Exception as e:
				logging.error(f'Failed to get avatar: {e}, {url}')
				return current

		if oldfilepath and isfile(oldfilepath) and filename != current:
			remove(oldfilepath)

		return filename


	def AddAcct(self, domain, apikey):
		userdata = self._request('api/v1/accounts/verify_credentials', domain=domain, apikey=apikey)
		instance = self._request('api/v1/instance', domain=domain, apikey=apikey)
		avatar = self.GetAvatar(userdata['avatar_static'])
		fullname = f'{userdata["username"]}@{domain}'.lower()

		data = {
			'displayname': userdata['display_name'],
			'username': userdata['username'],
			'domain': domain,
			'tootlimit': int(instance.get('max_toot_chars', 500)),
			'apikey': apikey,
			'fullname': fullname,
			'avatar': avatar
		}

		put.mastoaccount(data)

		return fullname


	def Login(self, domain, email, password):
		data = {
			'client_name': 'QtWeb',
			'redirect_uris': 'urn:ietf:wg:oauth:2.0:oob',
			'scopes': 'read write follow push',
			'website': 'https://git.barkshark.xyz/izaliamae/qtweb'
		}

		appid = self._request('api/v1/apps', method='POST', data=data, domain=domain)

		if not appid:
			logging.error('Did not receive client id and code')
			return

		newdata = {
			'redirect_uri': data['redirect_uris'],
			'scope': data['scopes'],
			'grant_type': 'password',
			'client_id': appid['client_id'],
			'client_secret': appid['client_secret'],
			'username': email,
			'password': password
		}

		tokendata = self._request('oauth/token', method='POST', data=newdata, domain=domain)

		if not tokendata:
			logging.error('Failed to login to mastodon')
			return False

		fullname = self.AddAcct(domain, tokendata['access_token'])
		return db.fetch('mastodon', fullname=fullname)


	def GetUser(self, fullname=None, row=None):
		row = db.fetch('mastodon', fullname=fullname) if not row else row
		return self._request('api/v1/accounts/verify_credentials', domain=row.domain, apikey=row.apikey)


	def RefreshData(self, fullname=None, row=None):
		row = db.fetch('mastodon', fullname=fullname) if not row else row

		if not fullname:
			if row:
				fullname = row.fullname

			else:
				logging.error('Full username not supplied')
				return

		if not row:
			logging.error(f'Cannot find user: {fullname}')
			return

		userdata = self.GetUser(self, row=row)

		if not userdata:
			logging.error(f'Failed to fetch user data for {fullname}. Was the token revoked?')
			return

		avatar = self.GetAvatar(userdata['avatar_static'], current=row.avatar)

		data = {
			'displayname': userdata['display_name'],
			'username': userdata['username'],
			'fullname': fullname,
			'avatar': avatar,
			'domain': row.domain
		}

		put.mastoaccount(data)


	def RefreshAll(self):
		for row in db.fetch('mastodon', single=False):
			self.RefreshData(row=row)


	def Post(self, fullname=None, warning=None, content=None, privacy=None, reply=None):
		row = db.fetch('mastodon', fullname=fullname)

		if not content:
			logging.error('Missing main toot content. Refusing to post.')
		data = {
			'visibility': privacy,
			'status': content
		}

		if warning:
			data['spoiler_text'] = warning

		if reply:
			data['in_reply_to_id'] = reply

		if not row:
			logging.error('User not found:', fullname)

		return self._request('api/v1/statuses', domain=row.domain, apikey=row.apikey, method='POST', data=data)


	def PostAction(self, action, url, fullname):
		row = get.mastoaccount(fullname)

		if action not in ['favourite', 'reblog', 'bookmark']:
			logging.error('Invalid post action:', action)

		if not row:
			logging.error('Invalid user:', fullname)
			return

		post = self.GetPostData(url, fullname)
		postid = post.get('id')

		if not post or post.get('error'):
			logging.error('Could not find post')
			return

		return self._request(f'api/v1/statuses/{postid}/{action}', domain=row.domain, apikey=row.apikey, method='POST')


	def GetPostData(self, url, fullname):
		row = db.fetch('mastodon', fullname=fullname)

		if not row:
			logging.error('Invalid user:', fullname)
			return

		search = self._request('api/v2/search', domain=row.domain, apikey=row.apikey, method='GET', data={'q': url, 'resolve': True, 'type': 'statuses'})

		if not search:
			return

		results = search['statuses']

		if len(results) > 0:
			return results[0]


	def Search(self, apikey, query):
		return self._request(apikey, 'api/v2/search', 'GET', {'q': query, 'type': 'statuses'})
