import asyncio, socket, json, sys, traceback, tracemalloc, argparse, datetime, linecache

from os.path import join
from pathlib import Path
from xml.sax.saxutils import unescape
from urllib.parse import urlparse

from . import isWindows
from .database import get, db
from .functions import ParseSoundcloudInfo
from .config import var

from .Lib.IzzyLib import logging
from .Lib.IzzyLib.misc import boolean

## QEventLoop is broken on windows
from asyncqt import QSelectorEventLoop


if logging.getConfig('level') == 'DEBUG':
	tracemalloc.start()


class Server(object):
	def __init__(self, window, sock=False):
		self.window = window
		self.GetWebview = self.window.ui.tabs.currentWidget
		self.socket = Path(join(var.profilepath, 'mbus.sock')) if sock else False
		self.loop = None
		self.server = None
		self.scinfo = None

		#todo: specify argument names
		self.commands = {
			'NewTab': self.new_tab,
			'CloseTab': self.close_tab,
			'ListTabs': self.gettabs,
			'Present': self.present,
			'Search': self.search,
			'Soundcloud': self.soundcloud,
			'Trace': self.trace,
			'Quit': self.quit,
			'CmdList': self.cmdlist,
			'Socks': self.socks
		}


	def start(self):
		self.loop = QSelectorEventLoop(self.window.app)

		args = [get.config('mhost', '127.0.0.1'), int(get.config('mport', 8008))]
		options = {'reuse_port': True, 'reuse_address': True} if not isWindows else {}

		if self.socket:
			self.server = asyncio.start_unix_server(self._client_handler, self.socket)
			address = self.socket

		else:
			self.server = asyncio.start_server(self._client_handler, *args, **options)
			address = ':'.join(str(arg) for arg in args)

		self.loop.create_task(self.server)

		if self.socket:
			## Not sure why the socket needs to be chmod'ed in a loop, but it works
			self.loop.create_task(self.set_perms())

		logging.verbose('Starting mbus at', address)


	async def set_perms(self):
		self.socket.chmod(0o600)


	def stop(self):
		if self.loop:
			self.loop.stop()
			self.loop.close()

		if self.socket:
			sock = Path(self.socket)

			if sock.exists():
				sock.unlink()


	def restart(self):
		if self.loop:
			self.stop()
		self.start()


	def format_cache(self, table=None):
		resp = {}

		for item in db.cache:
			tabledata = db.cache[item]

			if table and item != table:
				continue

			resp[item] = []

			for _, row in tabledata.items():
				if row:
					resp[item].append(row.asdict())

		return resp


	async def soundcloud(self, action):
		action = action.lower()
		webview = await self.get_soundcloud_tab_widget()

		if not webview:
			return False

		if action == 'info':
			webview.JsAction.SoundcloudInfo(self.get_soundcloud_info)

			await asyncio.sleep(0.1)

			text = self.scinfo.copy() if self.scinfo else None
			self.scinfo = None
			return text

		return webview.JsAction.SoundcloudPlayer(action)


	async def get_soundcloud_tab_widget(self):
		tabs = await self.gettabs()

		for tabid, data in tabs.items():
			if urlparse(data['url']).netloc == 'soundcloud.com':
				return self.window.ui.tabs.widget(int(tabid))


	def get_soundcloud_info(self, value):
		self.scinfo = ParseSoundcloudInfo(value)


	async def search(self, *args):
		searchbar = self.GetWebview().searchbar

		if not args:
			searchbar.Search('reset')
			searchbar.hide()
			return

		searchbar.Search(text=' '.join(args))
		searchbar.show()


	async def new_tab(self, *args, **kwargs):
		self.window.NewWebTab(*args, **kwargs)


	async def close_tab(self, index):
		self.window.CloseTab(index)


	async def gettabs(self):
		data = {}
		tabrange = list(range(0, self.window.ui.tabs.count()))

		for tabid in ([0] if len(tabrange) < 1 else tabrange):
			widget = self.window.ui.tabs.widget(tabid)
			if widget.windowTitle():
				continue

			data[tabid] = {'title': widget.title(), 'url': widget.url().toString()}

		return data


	async def present(self):
		self.window.ui.activateWindow()


	async def cmdlist(self):
		return list(self.commands.keys())


	async def trace(self, action, num=15):
		action = action.lower()
		actions = {
			'start': tracemalloc.start,
			'stop': tracemalloc.stop,
			'snap': self.trace_snapshot,
			'compare': self.trace_compare,
			'top': self.trace_top
		}

		cmd = actions.get(action)
		opts = {'num': num} if action in ['snap', 'compare'] else {}

		if not cmd:
			return

		return cmd(**opts)


	def trace_format(self, stats, num):
		data = ''

		for i, stat in enumerate(stats[:int(num)], 1):
			data += f'{i} {str(stat)}\n'

		return data


	def trace_snapshot(self, num=15):
		current = tracemalloc.take_snapshot()
		self.window.snapshot = current
		return self.trace_format(current.statistics('filename'), num)


	def trace_compare(self, num=15):
		current = tracemalloc.take_snapshot()

		if self.window.snapshot:
			stats = current.compare_to(self.window.snapshot, 'filename')

		else:
			stats = current.statistics('filename')

		return self.trace_format(stats, num)


	def trace_top(self, num=15, key_type='lineno'):
		if not self.window.snapshot:
			self.trace_snapshot()

		snapshot = self.window.snapshot.filter_traces((
			tracemalloc.Filter(False, '<frozen importlib._bootstrap>'),
			tracemalloc.Filter(False, '<unknown>'),
		))
		top_stats = snapshot.statistics(key_type)
		result = []

		result.append(f"Top {num} lines")

		for index, stat in enumerate(top_stats[:num], 1):
			frame = stat.traceback[0]
			result.append(f'#{index}: {frame.filename}:{frame.lineno}: {stat.size / 1024} KiB')
			line = linecache.getline(frame.filename, frame.lineno).strip()

			if line:
				result.append(f'    {line}')

		other = top_stats[num:]

		if other:
			size = sum(stat.size for stat in other)
			result.append(f'{len(other)} other: {size / 1024 / 1024} MiB')

		total = sum(stat.size for stat in top_stats)
		result.append(f'Total allocated size: {total / 1024 / 1024} MiB')

		return '\n'.join(result)


	async def socks(self):
		await self.present()
		self.window.NewWebTab('https://barkshark.xyz/web/statuses/103155447990974282', True)


	async def quit():
		self.window.app.quit()


	async def _client_handler(self, reader, writer):
		request = None

		while request != 'quit':
			try:
				reqdata = (await reader.read(81920)).decode('utf8')
				request = json.loads(reqdata)
				response = {}

				cmd = request.get('cmd')
				args = request.get('args', [])
				kwargs = request.get('kwargs', {})

				if not cmd:
					response = {'err': 'No command specified'}
					writer.write(f'{json.dumps(response)}\n'.encode('utf8'))
					continue

				command = self.commands.get(cmd)

				if not command:
					response = {'err': 'Invalid command'}
					writer.write(f'{json.dumps(response)}\n'.encode('utf8'))
					continue

				if cmd != 'CmdList':
					logging.debug(f'Running mbus command with args: {cmd}, args: {args}, kwargs: {kwargs}')

				output = await command(*args, **kwargs)

				if type(output) == dict:
					response = {'msg': output}

				elif output == False:
					response = {'err': 'Server error'}

				elif output == None:
					response = {'msg': 'OK'}

				elif output:
					response['msg'] = output

			except json.decoder.JSONDecodeError as e:
				if reqdata == '':
					response = {}
				else:
					logging.error('MBus JSON error:', repr(reqdata))
					response = {'err': f'JSON error: {str(e)}'}

			except Exception as e:
				traceback.print_exc()
				exc_type, value, tb = sys.exc_info()
				response = {'err': f'{exc_type.__name__}: {value}'}

			try:
				writer.write(f'{json.dumps(response)}\n'.encode('utf8'))
				await writer.drain()

			except ConnectionResetError:
				break

		writer.close()


class Client(object):
	def __init__(self, host=None, port=None, sock=None):
		self.socket = Path(sock) if type(sock) == str else sock
		self.conn = (socket.gethostbyname(host), port) if not sock else str(self.socket)


	def send(self, cmd, *args, **kwargs):
		try:
			socktype = socket.AF_INET if not self.socket else socket.AF_UNIX
			s = socket.socket(socktype, socket.SOCK_STREAM)
			s.settimeout(3)
			s.connect(self.conn)

		except socket.timeout:
			logging.error('Socket timeout. Is the server running?')
			return

		except (ConnectionRefusedError, FileNotFoundError):
			logging.verbose('Failed to connect to socket')
			return

		data = {'cmd': cmd, 'args': args, 'kwargs': kwargs}

		try:
			s.sendall(json.dumps(data).encode('utf8'))
			response = json.loads(s.recv(81920).decode('utf8'))

		except ConnectionResetError:
			logging.verbose('Lost mbus client connection')
			return

		s.close()

		return response


	def addcmd(self, cmd):
		setattr(self, cmd, lambda *args, **kwargs: self.send(cmd, *args, **kwargs))


	def setup(self):
		if self.socket and not self.socket.exists():
			logging.verbose('Unix socket does not exist')
			return False

		commands = self.send('CmdList')

		if not commands or not commands.get('msg'):
			return

		for cmd in commands.get('msg'):
			self.addcmd(cmd)

		return True


## Basic mbus client. Run via `python3 -m qtweb.mbus <cmd> <args>`
def main(args=sys.argv[1:]):
	#cmd = sys.argv[1] if len(sys.argv) > 1 else None
	#args = sys.argv[2:] if len(sys.argv) > 2 else []

	parser = argparse.ArgumentParser()
	parser.add_argument('cmd', help='The command to run')
	parser.add_argument('args', nargs=argparse.REMAINDER, help='The arguments to pass to the command')
	parser.add_argument('-a', '--address', help='Address to connect to. Default: localhost')
	parser.add_argument('-p', '--port', help='Port number to connect to. Default: 8008')
	parser.add_argument('-e', '--profile', help='Profile to use for a unix socket connection')
	pargs = parser.parse_args(args)

	host = '127.0.0.1' if not pargs.address else pargs.address
	port = int(8008 if not pargs.port else pargs.port)

	args = []
	kwargs = {}

	if pargs.cmd == 'NewTab':
		args = pargs.args

	else:
		for arg in pargs.args:
			splitarg = arg.split('=')

			if len(splitarg) == 1:
				args.append(arg)

			else:
				kwargs[splitarg[0]] = boolean(splitarg[1], fail=False)

	if not isWindows and any([pargs.address, pargs.port]):
		api = Client(host, port)

	else:
		profilepath = Path(var.datapath).joinpath('Default' if not pargs.profile else pargs.profile)
		api = Client(sock=profilepath.joinpath('mbus.sock'))

	if not api.setup():
		logging.error('Failed to connect to browser')
		sys.exit()

	try:
		command = getattr(api, pargs.cmd)
	except AttributeError:
		logging.error('Command not understood by server:', pargs.cmd)
		sys.exit()

	output = command(*args, **kwargs)

	if output:
		msg = output.get('msg')
		err = output.get('err')

		if err:
			logging.error('Server error:', err)

		if msg:
			logging.info(msg)

	else:
		logging.error('No output')


if __name__ == '__main__':
	main()
