import re, json

from pathlib import Path
from os.path import join, isfile, abspath
from urllib.parse import urlparse
from configparser import ConfigParser

from . import isWindows
from .config import var, default
from .database import db, put, get, delete
from .functions import Icon, NewDesktopEntry
from .dialogs import EditBookmark

from .Lib.IzzyLib import logging
from .Lib.IzzyLib.misc import boolean

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5 import uic


class Bookmarks(QWidget):
	def __init__(self, parent):
		super().__init__(parent)

		self.ui = uic.loadUi(join(var.resources, 'tab-bookmarks.ui'))
		self.setLayout(self.ui.layout())
		self.window = parent

		self.ui.add.clicked.connect(lambda *args: self._row_action('new'))
		self.ui.edit.clicked.connect(lambda *args: self._row_action('edit'))
		self.ui.open.clicked.connect(lambda *args: self._row_action('open'))
		self.ui.refresh.clicked.connect(lambda *args: self._row_action('refresh'))
		self.ui.table.cellDoubleClicked.connect(lambda *args: self._row_action('open'))

		for size in [(0, 350), (1, 200), (2, 80)]:
			self.ui.table.setColumnWidth(*size)

		self._populate_table()


	def _row_action(self, button):
		tablerow = self.ui.table.currentRow()

		if button not in ['new', 'refresh']:
			if tablerow == -1:
				return

			data = {
				'name': self.ui.table.item(tablerow, 0).text() if self.ui.table.item(tablerow, 0) else None,
				'url': self.ui.table.item(tablerow, 1).text() if self.ui.table.item(tablerow, 1) else None,
				'category': self.ui.table.item(tablerow, 2).text() if self.ui.table.item(tablerow, 2) else None,
				'description': self.ui.table.item(tablerow, 3).text() if self.ui.table.item(tablerow, 3) else None
			}

		else:
			data = {}

		actions = {
			'open': {'cmd': self.window.NewWebTab, 'args': [data.get('url')], 'kwargs': {'switch': True}},
			'new': {'cmd': lambda *args, **kwargs: EditBookmark(*args, **kwargs).ui.show()},
			'refresh': {'cmd': self._refresh_table},
			'edit': {
				'cmd': lambda *args, **kwargs: EditBookmark(*args, **kwargs).ui.show(),
				'kwargs': {
					'row': db.get('bookmarks', url=data.get('url'))
				}
			}
		}

		action = actions.get(button)

		if not action:
			return

		action['cmd'](*action.get('args', []), **action.get('kwargs', {}))


	def _populate_table(self):
		for row in db.fetch('bookmarks', single=False):
			table = self.ui.table
			rowpos = table.rowCount()

			table.insertRow(rowpos)
			table.setItem(rowpos, 0, QTableWidgetItem(row.name))
			table.setItem(rowpos, 1, QTableWidgetItem(row.url))
			table.setItem(rowpos, 2, QTableWidgetItem(row.category))
			table.setItem(rowpos, 4, QTableWidgetItem(row.description))


	def _refresh_table(self, *args):
		self.ui.table.setRowCount(0)
		self._populate_table()


	def _row_updated(self, *args):
		'maybe later'


class History(QWidget):
	def __init__(self, parent):
		self.window = parent


	def CloseTab(self):
		del self.window.widgettabs.history


## I'll get this working later
class Downloads(QWidget):
	def __init__(self, parent):
		super().__init__(parent)

		self.ui = uic.loadUi(join(var.resources, 'tab-downloads.ui'))
		self.setLayout(self.ui.layout())
		self.setWindowTitle('Downloads')
		self.window = parent


	def populate_table():
		self.ui.downloads.setRowCount(0)

		for download, md5 in self.window.profile.downloads:
			listItem = QListWidgetItem()
			dlItem = DownloadItem(self, download, md5)

			self.ui.downloads.addItem(listItem)
			self.ui.downloads.setItemWidget(listItem, dlItem)


	def CloseTab(self):
		del self.window.widgettabs.downloads


class Preferences(QWidget):
	def __init__(self, parent):
		super().__init__(parent)
		self.ui = uic.loadUi(join(var.resources, 'tab-preferences.ui'))
		self.setLayout(self.ui.layout())
		self.window = parent

		dldir = get.config('dlpath', var.downloadpath)
		nitter_state = get.config('nitter', 0)

		## General tab
		self.ui.homepage.setText(get.config('homepage', 'http://ddg.gg'))
		self.ui.downloaddir.setText(dldir)
		self.ui.adblock.setChecked(int(get.config('adblock', 1)))
		self.ui.hidetabs.setChecked(int(get.config('hidetabs', 1)))
		self.ui.reusetab.setChecked(int(get.config('reusetab', False)))
		self.ui.cachesize.setValue(int(get.config('cache', 100)))
		self.ui.nitter.setChecked(int(nitter_state))
		self.ui.nitterurl.setText(get.config('nitterurl', 'https://nitter.net'))
		self._general_set_nitterurl_state(nitter_state)

		if isWindows:
			self.ui.setdefault.setDisabled(True)

		if Path(dldir).exists():
			self.ui.createdownloaddir.hide()

		## Search engine tab
		for size in [(0, 25), (1, 200), (2, 65)]:
			self.ui.sengines.setColumnWidth(*size)

		## Mastodon tab
		for size in [(0, 25), (1, 300), (2, 150)]:
			self.ui.accounts.setColumnWidth(*size)

		## Advanced tab
		hostindex = self.ui.mbushost.findText(str(get.config('mhost', '127.0.0.1')))
		logindex = self.ui.loglevel.findText(get.config('loglevel', 'Info').title())
		self.ui.mbushost.setCurrentIndex(hostindex)
		self.ui.mbusport.setValue(int(get.config('mport', 8008)))
		self.ui.loglevel.setCurrentIndex(logindex)
		self.ui.jsdebug.setChecked(get.config('jsdebug', False))
		self.ui.tracemalloc.setChecked(get.config('tracestart', False))
		self.ui.automd5.setChecked(get.config('automd5', False))
		self.ui.mastocheck.setChecked(get.config('mastocheck', False))

		## General tab signals
		self.ui.setdefault.clicked.connect(self._general_set_default)
		self.ui.selectdownloaddir.clicked.connect(self._general_pick_download_dir)
		self.ui.createdownloaddir.clicked.connect(self._general_create_download_dir)
		self.ui.downloaddir.returnPressed.connect(self._general_pick_download_dir)
		self.ui.downloaddir.editingFinished.connect(self._general_set_download_dir)
		self.ui.homepage.editingFinished.connect(self._general_update_homepage)
		self.ui.cachesize.editingFinished.connect(self._general_update_cachesize)
		self.ui.hidetabs.stateChanged.connect(lambda state: self._general_handle_checkbox(state, 'hidetabs'))
		self.ui.adblock.stateChanged.connect(lambda state: self._general_handle_checkbox(state, 'adblock'))
		self.ui.nitter.stateChanged.connect(lambda state: self._general_handle_checkbox(state, 'nitter'))
		self.ui.reusetab.stateChanged.connect(lambda state: self._general_handle_checkbox(state, 'reusetab'))
		self.ui.nitterurl.editingFinished.connect(self._general_update_nitter_url)
		self.ui.nitterreset.clicked.connect(self._general_reset_nitter_url)

		## Appearance tab signals
		#self.ui.theme.itemClicked.connect(self._appearance_theme_change)

		## Search engine tab signals
		self.ui.snew.clicked.connect(lambda *args: self._search_button_handler('new'))
		self.ui.sdelete.clicked.connect(lambda *args: self._search_button_handler('delete'))
		self.ui.sdefault.clicked.connect(lambda *args: self._search_button_handler('default'))
		self.ui.sengines.cellChanged.connect(self._search_edit_cell)

		## Mastodon tab signals
		self.ui.login.clicked.connect(self.mastodon_login)
		self.ui.reset.clicked.connect(self.mastodon_reset)
		self.ui.logout.clicked.connect(self.mastodon_logout)
		self.ui.mdefault.clicked.connect(self.mastodon_set_default)
		self.ui.refresh.clicked.connect(self.refresh_accounts)

		self.ui.domain.returnPressed.connect(lambda *args: self.ui.username.setFocus())
		self.ui.username.returnPressed.connect(lambda *args: self.ui.password.setFocus())
		self.ui.password.returnPressed.connect(self.mastodon_login)
		self.ui.accounts.itemSelectionChanged.connect(self.mastodon_row_selected)

		## Advanced tab signals
		self.ui.mbushost.currentIndexChanged.connect(self._adv_mbushost_changed)
		self.ui.mbusport.editingFinished.connect(self._adv_mbusport_changed)
		self.ui.loglevel.currentIndexChanged.connect(lambda index: self._adv_changed('loglevel', index))
		self.ui.hostreset.clicked.connect(lambda *args: self._adv_mbushost_changed(int(default.mhost)))
		self.ui.portreset.clicked.connect(lambda *args: self._adv_mbusport_changed(default.mport))
		self.ui.logreset.clicked.connect(lambda *args: self._adv_reset('loglevel'))
		self.ui.jsdebug.stateChanged.connect(lambda state: self._adv_handle_checkbox('jsdebug', state))
		self.ui.tracemalloc.stateChanged.connect(lambda state: self._adv_handle_checkbox('tracestart', state))
		self.ui.automd5.stateChanged.connect(lambda state: self._adv_handle_checkbox('automd5', state))
		self.ui.mastocheck.stateChanged.connect(lambda state: self._adv_handle_checkbox('mastocheck', state))

		## Populate tables
		self.mastodon_populate_table()
		self.populate_search_table()
		#self._appearance_populate_view()

		# Remove the appearance tab until I can get it fixed
		self.ui.tabs.widget(1).deleteLater()
		self.ui.tabs.removeTab(1)


	### General Tab ###
	def _general_set_default(self):
		if isWindows:
			'heck'

		else:
			applications = join(var.home, '.local', 'share', 'applications')
			mimeconfig = ConfigParser()
			mimelist = None
			newfile = True
			mimeapps = [
				Path(join(var.home, '.config', 'mimeapps.list')),
				Path(join(applications, 'mimeapps.list'))
			]

			for mimeapp in mimeapps:
				if mimeapp.exists():
					mimelist = mimeapp
					break

			if not mimelist:
				mimelist = mimeapps[0]

			else:
				newfile = False

			if not newfile:
				mimeconfig.read(mimelist)
			else:
				mimeconfig['Default Applications'] = {}

			deskfile = NewDesktopEntry(Path(applications))

			newmimelist = {
				'application/x-extension-htm': deskfile,
				'application/x-extension-html': deskfile,
				'application/x-extension-shtml': deskfile,
				'application/x-extension-xht': deskfile,
				'application/xhtml+xml': deskfile,
				'application/xml': deskfile,
				'text/html': deskfile,
				'x-scheme-handler/ftp': deskfile,
				'x-scheme-handler/http': deskfile,
				'x-scheme-handler/https': deskfile,
				'x-scheme-handler/about': deskfile,
				'x-scheme-handler/chrome': deskfile,
				#'x-scheme-handler/qtweb': deskfile,
			}

			mimeconfig['Default Applications'].update(newmimelist)
			mimeconfig['Added Associations'].update(newmimelist)

			with mimelist.open('w') as fd:
				mimeconfig.write(fd)


	def _general_pick_download_dir(self):
		dirname = QFileDialog.getExistingDirectory(caption='Select Download Directory', directory=var.home)

		if dirname:
			self._general_set_download_dir(dirname)


	def _general_set_download_dir(self, path=None):
		if type(path) == Path:
			dlpath = path

		else:
			dlpath = Path(self.ui.downloaddir.text() if not path else path)

		if not dlpath.exists():
			self.ui.createdownloaddir.show()
			return

		elif not dlpath.is_dir():
			self.window.notif.Error('Download path is a file. Not updating')
			return

		else:
			self.ui.createdownloaddir.hide()

		if not put.config('dlpath', str(dlpath)):
			self.window.notif.Error('Failed to update download directory')

		else:
			self.ui.downloaddir.setText(str(dlpath))


	def _general_create_download_dir(self):
		dlpath = Path(self.ui.downloaddir.text())
		dlpath.mkdir(parents=True, exist_ok=True)

		self._general_set_download_dir(dlpath)


	def _general_update_cachesize(self, *args):
		value = int(self.ui.cachesize.value())
		put.config('cache', value)
		self.window.profile.setHttpCacheMaximumSize(1024 * 1024 * value)


	def _general_update_homepage(self, *args):
		newurl = self.ui.homepage.text()

		if urlparse(newurl).scheme == '':
			newurl = 'https://' + newurl

		put.config('homepage', newurl)
		self.window.homepage = newurl


	def _general_handle_checkbox(self, state, name):
		value = 1 if state.real == 2 else 0
		put.config(name, str(value))

		if name == 'hidetabs':
			self.window.tabbar.setAutoHide(state.real)

		if name == 'nitter':
			value = boolean(value)
			self._general_set_nitterurl_state(value)


	def _general_set_nitterurl_state(self, value=None):
		if not value:
			value = get.config('nitter', False)

		self.ui.nitterurl.setEnabled(value)
		self.ui.nitterreset.setEnabled(value)


	def _general_update_nitter_url(self):
		url = self.ui.nitterurl.text()

		if not any(map(url.startswith, ['http://', 'https://'])):
			url = 'https://' + url
			self.ui.nitterurl.setText(url)

		if not put.config('nitterurl', url):
			logging.error('Failed to update "nitterurl" config')


	def _general_reset_nitter_url(self):
		if not put.config('nitterurl', default.nitter):
			logging.error('Failed to update "nitterurl" config')
			return

		self.ui.nitterurl.setText(default.nitter)


	### Appearance Tab ###
	def _appearance_populate_view(self):
		self.ui.theme.clear()
		size = QSize()
		size.setHeight(24)

		default = QListWidgetItem('Default')
		default.setSizeHint(size)

		if get.config('theme', 'Default') == 'Default':
			default.setIcon(Icon('okay', 16))

		self.ui.theme.addItem(default)

		for theme in get.theme():
			item = QListWidgetItem(theme.name)
			item.setForeground(theme.palette.text())
			item.setBackground(theme.palette.base())
			item.setSizeHint(size)

			if get.config('theme') == theme.name:
				item.setIcon(Icon('okay', 16))

			self.ui.theme.addItem(item)


	def _appearance_theme_change(self, item):
		listwidget = self.ui.theme
		theme = item.text()

		if put.config('theme', theme):
			items = [listwidget.item(i) for i in range(listwidget.count())]

			for item in items:
				if item.text() != theme:
					item.setIcon(QIcon())
				else:
					item.setIcon(Icon('okay', 16))

			self.window.themes.ChangeTheme(theme)


	### Search Tab ###
	def _search_button_handler(self, button):
		table = self.ui.sengines
		tablerow = table.currentRow()
		cells = [0, table.item(tablerow, 1), table.item(tablerow, 2), table.item(tablerow, 3)]

		data = {
			'name': cells[1].text() if cells[1] else None,
			'keyword': cells[2].text() if cells[2] else None,
			'url': cells[3].text() if cells[3] else None
		}

		if button in ['edit', 'delete'] and None in data.values():
			logging.debug('No data')
			return

		actions = {
			'new': {'cmd': self.ui.sengines.insertRow, 'args': [self.ui.sengines.rowCount()]},
			'edit': {'cmd': print, 'args': [data]},
			'delete': {'cmd': delete.search, 'args': [data['keyword']]},
			'default': {'cmd': self.search_set_default, 'args': [data]}
		}

		action = actions.get(button)

		if not action:
			logging.error('Invalid button:', button)
			return

		if action['cmd'](*action.get('args', []), **action.get('kwargs', {})):
			if button == 'delete':
				# figure out how to delete a single row instead of refreshing the whole table
				self.populate_search_table()


	def _search_edit_cell(self, x, y):
		if self.update:
			return

		table = self.ui.sengines
		cells = [0, table.item(x, 1), table.item(x, 2), table.item(x, 3)]

		data = {
			'name': cells[1].text() if cells[1] else None,
			'keyword': cells[2].text() if cells[2] else None,
			'url': cells[3].text() if cells[3] else None
		}

		if None in data.values():
			logging.verbose('Refusing to update search engine with empty values')
			return

		logging.verbose('Updated search engine' if put.search(data) else 'Failed to update search engine')


	def search_set_default(self, data):
		if not data.get('keyword'):
			logging.error('Missing keyword:', data)
			return

		put.config('search', data.get('keyword'))
		self.populate_search_table()


	def populate_search_table(self):
		self.update = True
		table = self.ui.sengines
		table.setRowCount(0)

		for row in db.fetch('search', single=False, sort='keyword'):
			rowpos = table.rowCount()
			default = QTableWidgetItem()

			table.insertRow(rowpos)

			if row.keyword == get.config('search'):
				default.setIcon(Icon('okay'))

			table.setItem(rowpos, 0, default)
			table.setItem(rowpos, 1, QTableWidgetItem(row.name))
			table.setItem(rowpos, 2, QTableWidgetItem(row.keyword))
			table.setItem(rowpos, 3, QTableWidgetItem(row.url))

		self.update = False


	### Mastodon tab ###
	def mastodon_display_message(self, *args, msgtype='info'):
		msg = ' '.join(args)

		if msgtype == 'info':
			self.window.notif.Info(msg)

		elif msgtype == 'error':
			self.window.notif.Error(msg)

		else:
			logging.error('Invalid message type:', msgtype)


	def mastodon_row_selected(self):
		row = self.mastodon_selected_row_data()
		logging.debug(row)

		if not row:
			self.ui.mdefault.setEnabled(False)
			self.ui.logout.setEnabled(False)

		else:
			self.ui.mdefault.setEnabled(row['default'])
			self.ui.logout.setEnabled(True)


	def mastodon_selected_row_data(self):
		table = self.ui.accounts
		tablerow = table.currentRow() if table.currentRow() >= 0 else None

		if tablerow == None:
			return

		cells = [table.item(tablerow, 0), table.item(tablerow, 1), table.item(tablerow, 2), table.item(tablerow, 3)]

		data = {
			'default': cells[0].icon().isNull(),
			'displayname': cells[1].text() if cells[1] else None,
			'username': cells[2].text() if cells[2] else None,
			'domain': cells[3].text() if cells[3] else None
		}

		data['fullname'] = data['username'] + '@' + data['domain']
		return data


	def mastodon_populate_table(self):
		self.ui.accounts.setRowCount(0)
		accounts = get.mastoaccount()

		for widget in [self.ui.mdefault, self.ui.logout]:
			widget.setEnabled(False)

		#for row in db.get('mastodon', single=False):
		for row in accounts:
			self.mastodon_insert_row(row)


	def mastodon_insert_row(self, row):
		table = self.ui.accounts
		rowpos = table.rowCount()
		iconpath = join(var.profilepath, 'avatars', row.avatar) if row.avatar else None

		dispname = [row.displayname]
		default = QTableWidgetItem()

		table.insertRow(rowpos)

		if row.fullname == get.config('defaccount'):
			default.setIcon(Icon('okay'))

		if iconpath and isfile(iconpath):
			dispname.insert(0, QIcon(iconpath))

		table.setItem(rowpos, 0, default)
		table.setItem(rowpos, 1, QTableWidgetItem(*dispname))
		table.setItem(rowpos, 2, QTableWidgetItem(row.username))
		table.setItem(rowpos, 3, QTableWidgetItem(row.domain))


	def mastodon_remove_row(self, index):
		'heck'


	def mastodon_login(self, *args):
		worker = MastodonLogin(tab=self)
		self.window.threadpool.start(worker)


	def mastodon_logout(self, *args):
		data = self.mastodon_selected_row_data()
		fullname = data['fullname']
		result = delete.mastoaccount(fullname)

		if result:
			self.window.notif.Info(f'Logged out of Mastodon account: {fullname}')

		elif result == None:
			self.window.notif.Error(f'Account cannot be found in the database: {fullname}')

		elif result == False:
			self.window.notif.Error(f'Error when removeing account from database: {fullname}')

		self.mastodon_populate_table()


	def mastodon_set_default(self, *args):
		data = self.mastodon_selected_row_data()

		if not data or not data.get('fullname'):
			self.mastodon_display_message('Failed to set default account', msgtype='error')
			return

		fullname = data['fullname'].lower()

		if not get.mastoaccount(fullname):
			self.mastodon_display_message('Account not logged in:', fullname, msgtype='error')
			return

		put.config('defaccount', fullname)
		self.mastodon_populate_table()


	def mastodon_reset(self, *args):
		for widget in [self.ui.domain, self.ui.username, self.ui.password]:
			widget.setText('')


	def refresh_accounts(self, *args):
		worker = RefreshAccounts(tab=self)
		self.window.threadpool.start(worker)


	## Advanced tab
	def _adv_mbushost_changed(self, index):
		new = self.ui.mbushost.itemText(index)
		old = get.config('mhost', default.mhost)

		if old != new:
			put.config('mhost', new)

			if new == 'False':
				self.window.server.stop()

			else:
				self.window.server.restart()


	def _adv_mbusport_changed(self, value=None):
		if not value:
			value = self.ui.mbusport.value()

		put.config('mport', value)

		self.window.server.restart()


	def _adv_changed(self, field, *args):
		if field == 'loglevel':
			text = self.ui.loglevel.itemText(args[0]).upper()
			put.config('loglevel', text)
			logging.setConfig({'level': text})


	def _adv_handle_checkbox(self, field, state):
		value = 1 if state.real == 2 else 0
		put.config(field, value)


	def _adv_reset(self, field):
		row = db.fetch('config', key=field)

		if not row:
			logging.error(f'Config "{field}" not found')
			return

		db.remove('config', row.id)

		if field == 'loglevel':
			self.mbushost.setCurrentIndex(3)
			logging.setConfig({'level': 'INFO'})


	def CloseTab(self):
		del self.window.widgettabs.preferences


class DownloadItem(QWidget):
	def __init__(self, tab, download, md5=None):
		self.ui = uic.loadUi(join(var.resources, 'download-item.ui'))
		self.setLayout(self.ui.layout())
		self.tab = tab
		self.finished = False

		location = Path(download.downloadDirectory()).joinpath(download.downloadFileName())
		url = download.url().toString()

		self.ui.location.setText(str(location))
		self.ui.url.setText(url)

		download.downloadProgress.connect(self._update_progress)


		def _update_progress(self, current, total):
			precent = current / total * 100
			current_mb = current / (1024 ** 2)
			total_mb = total / (1024 ** 2)

			for value in [percent, current_mb, total_mb]:
				if type(value) == float:
					value = round(value, 2)

			progress = f'{percent}% - {current_mb}MiB/{total_mb}MiB'
			self.ui.percent.setText(progress)


class RefreshAccounts(QRunnable):
	def __init__(self, *args, tab=None, **kwargs):
		super().__init__(*args, **kwargs)
		self.tab = tab

	def run(self):
		self.tab.mastodon_display_message('Refreshing account data')
		self.tab.window.mastodon.RefreshAll()
		self.tab.mastodon_populate_table()
		self.tab.mastodon_display_message(f'Finished refresing account data :3')


class MastodonLogin(QRunnable):
	def __init__(self, *args, tab=None, **kwargs):
		super().__init__(*args, **kwargs)
		self.tab = tab


	def run(self):
		newacct = False if db.count('mastodon') > 0 else True
		domain = self.tab.ui.domain
		username = self.tab.ui.username
		password = self.tab.ui.password

		self.tab.mastodon_display_message(f'Logging into account at {domain.text()}')
		row = self.tab.window.mastodon.Login(domain.text(), username.text(), password.text())

		if row:
			self.tab.mastodon_display_message(f'Successfully logged in {row.fullname}')

			domain.setText('')
			username.setText('')
			password.setText('')

			if newacct:
				put.config('defaccount', row.fullname)

			self.tab.mastodon_insert_row(row)

		elif row == False:
			self.tab.mastodon_display_message(f'Invalid username or password', msgtype='error')

		else:
			self.tab.mastodon_display_message(f'Network error', msgtype='error')
