import threading, sys, os

from datetime import datetime
from os.path import join
from pathlib import Path

from .Lib.IzzyLib import logging

from . import __application__, __version__, pyqtver, debstable, isWindows
from . import dialogs
from .config import var
from .database import db, get, put, delete
from .functions import useragent, GetSoftware, httpclient, Decode, psl, MD5Check, WebEngineScript, DotDict

from PyQt5.QtWebEngineWidgets import *
from PyQt5.QtCore import QUrl, QDateTime
from PyQt5.QtWebEngineCore import QWebEngineUrlRequestInterceptor
from PyQt5.QtNetwork import QNetworkCookie


AddHeader = lambda info, k, v: info.setHttpHeader(k.encode(), v.encode())
blockedDomain = [
	'doubleclick.net',
	'rubiconproject.com',
	'outbrain.com',
	'googletagservices.com',
	'amazon-adsystem.com',
	'optimizely.com',
	'adswizz.com',
	'scorecardsearch.com'
]

blockedURL = {
	'youtube.com': [
		'/api/stats/ads',
		'/pagead/adview',
		'/pagead/lvz'
		'/get_midroll/info'
	],
	'soundcloud.com': [
		'/audio-ads',
		'/me'
	],
	'facebook.com': [
		'/tr',
		'/brandlift.php'
	]
}


class WebEngineProfile(QWebEngineProfile):
	def __init__(self, parent):
		super().__init__(var.profile, parent)

		self.interceptor = WebFilter(parent)
		self.window = parent
		self.notif = parent.notif
		self.user_scripts = []
		self.cookies = []
		self.downloads = {}

		self.downloadRequested.connect(self._download_request)

		self.cookieStore().setCookieFilter(self._cookie_filter)
		self.setCachePath(var.profilepath)
		self.setHttpCacheMaximumSize(int(get.config('cache', 100))*1024*1024)
		self.setPersistentStoragePath(var.profilepath)
		self.setHttpUserAgent(f'{self.httpUserAgent()} {__application__}/{__version__}')
		self.setSpellCheckEnabled(True)
		self.setSpellCheckLanguages(['en-US'])

		try:
			self.setUrlRequestInterceptor(self.interceptor)
		except AttributeError:
			self.setRequestInterceptor(self.interceptor)

		if not debstable:
			self.setNotificationPresenter(self._handle_web_notification)
			self.setDownloadPath(get.config('dlpath', var.downloadpath))

		self.attrs = {
			#QWebEngineSettings.DnsPrefetchEnabled: True,
			QWebEngineSettings.FullScreenSupportEnabled: True,
			QWebEngineSettings.JavascriptCanOpenWindows: True,
			QWebEngineSettings.JavascriptEnabled: True,
			QWebEngineSettings.PlaybackRequiresUserGesture: True,
			QWebEngineSettings.ScrollAnimatorEnabled: True,
			QWebEngineSettings.Accelerated2dCanvasEnabled: True
		}

		for k, v in self.attrs.items():
			self.settings().setAttribute(k, v)

		self.LoadAllScripts()

		# Will enable this once I figure out how to properly handle cookies
		#self.setPersistentCookiesPolicy(QWebEngineProfile.NoPersistentCookies)
		#self.LoadCookies()
		#self.cookieStore().cookieAdded.connect(self._new_cookie)
		#self.cookieStore().cookieRemoved.connect(self._rem_cookie)


	def _handle_web_notification(self, notif):
		host = notif.origin().host()
		site = get.permissions(host)

		if site.notification:
			options = {
				'title': notif.title(),
				'message': notif.message()
			}

			imgformat = 'ico' if isWindows else 'png'
			image_path = join(var.temp, f'{host}.{imgformat}')

			if notif.icon().save(image_path, format=imgformat):
				options['icon'] = image_path

			self.window.notif.New(**options)


	def _download_request(self, download):
		self.downloads[download.id()] = DotDict({'item': download, 'md5': None})
		dldialog = dialogs.NewDownload(self, self.downloads[download.id()])
		dldialog.exec_()


	def _download_finished(self, download):
		if debstable:
			filename = Path(download.path())

		else:
			filename = Path(download.item.downloadDirectory(), download.item.downloadFileName())

		url = download.item.url().toString()

		if download.md5 and filename.exists() and MD5Check(filename, download.md5) == False:
			logging.debug(f'Download failed md5 check: {filename}')
			self.window.notif.New('Download failed md5 check', filename)
			filename.unlink()

		elif download.item.state() == QWebEngineDownloadItem.DownloadCompleted:
			logging.debug(f'Finished downloading {url} to {filename}')
			self.window.notif.New('Finished Download', str(filename))

		#if download.item.state() == QWebEngineDownloadItem.DownloadCancelled:
			#self.notif.New('Cancelled Download', filename)

		del self.downloads[download.item.id()]


	def _cookie_filter(self, request):
		if request.thirdParty:
			url = request.firstPartyUrl if not request.firstPartyUrl.scheme() == 'data' else request.origin
			logging.verbose('Rejected 3rd-party cookie from', url.toString())
			return False

		return True


	def GetScript(self, index):
		return self.user_scripts[index]


	def LoadScript(self, filename):
		script = Path(filename)
		webscript = WebEngineScript(self.scripts(), script, script.match(f'{var.scriptpath}/*'))
		webscript.enable()

		self.user_scripts.append(webscript)


	def UnloadScript(self, index):
		script = index if type(index) == WebEngineScript else self.user_scripts[index]

		return script.disable()


	def DeleteSCript(self, index):
		script = index if type(index) == WebEngineScript else self.user_scripts[index]

		if not script.user:
			logging.error('Refusing to unload a system script')

		script.disable()
		script.filename.unlink()

		self.user_scripts.remove(script)


	def LoadAllScripts(self):
		scripts = [s for s in Path(var.scriptpath, 'scripts').rglob('*.sys.js')]
		scripts += [s for s in Path(var.profilepath, 'Scripts').rglob('*.user.js')]

		for script in scripts:
			self.LoadScript(script)


	def LoadCookies(self):
		for cookie in get.cookie():
			self.cookieStore().setCookie(cookie)


class WebFilter(QWebEngineUrlRequestInterceptor):
	def __init__(self, window):
		super().__init__()
		self.window = window
		self.threads = {}

	def interceptRequest(self, info):
		requrl = info.requestUrl()
		reqdomain = requrl.host()
		reqtopdomain = psl.privatesuffix(reqdomain)
		reqpath = requrl.path()

		pageurl = info.firstPartyUrl()
		pagedomain = pageurl.host()
		pagetopdomain = psl.privatesuffix(pagedomain)
		pagepath = pageurl.path()

		pagerow = get.permissions(pagedomain, defaults=True)

		#if len(self.threads.keys()) > 1:
			#logging.debug('Threads:', self.threads)

		## Prevent secure pages from loading insecure content
		if not pagerow.allowhttp and requrl.scheme() == 'http' and pageurl.scheme() == 'https':
			logging.verbose('Blocked insecure content:', requrl.toString().split('?', 1)[0])
			info.block(True)
			return

		## Block requests if they're in the block list
		if get.config('adblock') and pagerow.adblock and (reqpath in blockedURL.get(reqtopdomain, []) or reqtopdomain in blockedDomain or reqdomain in blockedDomain):
			logging.verbose('Blocked content:', requrl.toString().split('?', 1)[0])
			info.block(True)
			return

		## Check if domain is a mastodon instance
		if get.config('mastocheck', False):
			if not pagerow.domain and not pagerow.mastodon and not any(map(requrl.toString().startswith, ['about:', 'chrome://'])):
				if not pagedomain:
					return

				if self.threads.get(pagedomain):
					return

				logging.verbose('Checking for presence of nodeinfo')
				self.threads[pagedomain] = threading.Thread(target=GetSoftware, args=[pagedomain])
				self.threads[pagedomain].start()

			## Delete the thread object once it's finished
			elif self.threads.get(pagedomain):
				del self.threads[pagedomain]

		## Set custom headers
		# This breaks gmx.com
		#AddHeader(info, 'X-Trans-Rights', 'Are Human Rights')
