from urllib.parse import urlparse, quote_plus
from pathlib import Path

import validators

from .Lib.IzzyLib import logging
from .Lib.IzzyLib.misc import boolean
from .Lib.IzzyLib.cache import LRUCache

from . import debstable, dialogs, debstable
from .config import var, default
from .database import db, get, put
from .mastodon import Mastodon
from .functions import Menu, Icon, httpclient
from .webviewjs import JsFuncs

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtWebEngineWidgets import *
from PyQt5 import uic


webactions = [
	QWebEnginePage.Reload,
	QWebEnginePage.Stop,
	QWebEnginePage.Back,
	QWebEnginePage.Forward,
	QWebEnginePage.Undo,
	QWebEnginePage.Redo,
	QWebEnginePage.Cut,
	QWebEnginePage.Copy,
	QWebEnginePage.Paste,
	QWebEnginePage.SelectAll,
	QWebEnginePage.ExitFullScreen
]


class WebEngineView(QWebEngineView):
	def __init__(self, window, url=None, text=None):
		super().__init__()

		self.window = window
		self.jscache = LRUCache(64)
		self.webview = True
		self.tabs = window.ui.tabs
		self.webpage = WebEnginePage(self.window.profile, self)
		self.devpage = DevPage(self.window.profile, self.window)
		self.searchbar = SearchBar(self)

		self.layout().addWidget(self.searchbar)

		self.statusurl = None
		self.currenttext = None
		self.isFullscreen = False

		self.setContextMenuPolicy(Qt.CustomContextMenu)
		self.customContextMenuRequested.connect(self._web_context_menu)
		self.loadProgress.connect(self._progress_loading)
		#self.loadStarted.connect(self._started_loading)
		self.loadFinished.connect(self._finished_loading)

		self.webpage.titleChanged.connect(self._set_tab_text)
		self.webpage.iconChanged.connect(self._set_tab_icon)
		self.webpage.urlChanged.connect(self._set_tab_url)
		self.webpage.linkHovered.connect(self._link_hover)
		self.webpage.fullScreenRequested.connect(self._fullscreen_request)
		self.webpage.featurePermissionRequested.connect(self._permission_request)
		self.webpage.authenticationRequired.connect(self._handle_auth)

		self.webaction_state = {
			'previous': False,
			'forward': False,
			'stop': False,
			'reload': False
		}

		for webaction in webactions:
			action = self.webpage.action(webaction)
			action.changed.connect(self._webaction_state_change)

		self.setPage(self.webpage)
		self.JsAction = JsFuncs(self)

		if url != 'newtab':
			self.LoadUrl(url if url else get.config('homepage', default.homepage))


	def _progress_loading(self, progress):
		current_index = self.window.ui.tabs.currentIndex()
		tab_index = self.window.ui.tabs.indexOf(self)
		loading = progress != 100
		percent = progress / 100

		self.webaction_state['stop'] = loading
		self.webaction_state['reload'] = not loading

		if current_index == tab_index:
			#if loading:
				#self.window.ui.url.setStyleSheet('background: qlineargradient(x1:0, y1:0, x2:1, y2:0, stop: 0 #1bb77b, stop: ' + str(percent) + ' #1bb77b, stop: ' + str(percent+ 0.001) + ' rgba(0, 0, 0, 0), stop: 1 white)')
			#else:
				#self.window.ui.url.setStyleSheet(None)

			self.window._update_webaction_buttons('stop', self.webaction_state['stop'])
			self.window._update_webaction_buttons('reload', self.webaction_state['reload'])


	def _finished_loading(self, arg):
		host = self.url().host()
		jsfile = Path(var.scriptpath, 'scripts', host + '.js')
		jscache = self.jscache.fetch(host)

		if jsfile.is_file():
			if not jscache:
				with jsfile.open('r') as fd:
					data = fd.read()
					self.jscache.store(host, data)
					jscache = data

			self.webpage.runJavaScript(jscache)


	def _set_tab_text(self, title):
		index = self.tabs.indexOf(self)
		self.tabs.setTabText(index, title)
		self.tabs.setTabToolTip(index, title)


	def _set_tab_icon(self, icon):
		index = self.tabs.indexOf(self)
		self.tabs.setTabIcon(index, icon)


	def _set_tab_url(self, url):
		self.currenttext = url.toString()
		self.devpage.setWindowTitle(f'QtWeb Inspector: {url.toString()}')
		index = self.tabs.indexOf(self)

		if index == self.tabs.currentIndex():
			url = self.currenttext
			self.window.ui.url.setText(url)

		self.window.ui.url.setCursorPosition(0)


	def _webaction_state_change(self):
		current_index = self.window.ui.tabs.currentIndex()
		tab_index = self.window.ui.tabs.indexOf(self)
		recv_action = self.sender()

		action = recv_action.text().lower()
		state = recv_action.isEnabled()

		if action in ['stop', 'reload']:
			return

		self.webaction_state[action] = state

		if current_index == tab_index:
			self.window._update_webaction_buttons(action, state)


	def _permission_request(self, origin, feature):
		host = origin.host()
		site = get.permissions(host)

		features = {
			QWebEnginePage.Geolocation: (site.location, 'Location'),
			QWebEnginePage.MediaAudioCapture: (site.microphone, 'Microphone'),
			QWebEnginePage.MediaVideoCapture: (site.camera, 'Webcam'),
			QWebEnginePage.MediaAudioVideoCapture: (site.microphone and site.camera, 'Microphone and Webcam'),
			QWebEnginePage.DesktopVideoCapture: (site.capture, 'Desktop Video Capture'),
			QWebEnginePage.DesktopAudioVideoCapture: (site.capture, 'Desktop Audio/Video Capture'),
		}

		if not debstable:
			features[QWebEnginePage.Notifications] = (site.notification, 'Notification')

		permission, name = features.get(feature, (None, 'Invalid Permission'))

		if permission:
			self.webpage.setFeaturePermission(origin, feature, QWebEnginePage.PermissionGrantedByUser)
			logging.verbose(f'Allowed permission for {host}:', name)

		else:
			self.webpage.setFeaturePermission(origin, feature, QWebEnginePage.PermissionDeniedByUser)
			logging.verbose(f'Rejected permission for {host}:', name)


	def _fullscreen_request(self, request):
		host = request.origin().host()
		site = get.permissions(host, defaults=True)

		if site.fullscreen:
			logging.verbose('Allowing fullscreen for', host)
			request.accept()
			self.Fullscreen(request.toggleOn())

		else:
			logging.verbose('Disallowing fullscreen for ', host)


	def _handle_auth(self, location, authenticator):
		url = location.toString()
		login = dialogs.HttpAuth(location, authenticator, self)
		login.exec_()

		if login.result:
			logging.verbose('Authenticated page:', url)

		else:
			logging.verbose('Canceled authentication for page:', url)
			self.stop()

			if self.window.ui.tabs.currentWidget() == self:
				self.window.ui.url.setText(url)


	def _web_context_menu(self, point):
		cursor = QCursor().pos()
		data = self.webpage.contextMenuData()
		clipboard = self.window.clipboard
		currenturl = self.url()
		spliturl = currenturl.path().split('/')
		selected = data.selectedText()
		search = get.search(get.config('search', 'ddg'))

		urlismasto = get.permissions(currenturl.host(), defaults=True).mastodon
		urlispost = (len(spliturl) == 3 and spliturl[1].startswith('@'))

		linktext = data.linkText()
		linkurl = data.linkUrl().toString()
		linksplit = data.linkUrl().path().split('/')
		linkismasto = get.permissions(urlparse(linkurl).netloc, defaults=True).mastodon if urlparse(linkurl).netloc != '' else None
		linkispost = (len(linksplit) == 3 and linksplit[1].startswith('@'))
		acctcount = db.count('mastodon') > 0

		mediaurl = data.mediaUrl().toString()
		menu = Menu(self)

		## Not sure if I wanna enable this
		#if acctcount and urlismasto and urlispost:
			#addItem('Reply to Post', 'reply', currenturl.toString())
			#addItem('Favorite Post', 'favourite', currenturl.toString())
			#addItem('Boost Post', 'reblog', currenturl.toString())
			#addItem('Bookmark Post', 'bookmark', currenturl.toString())

		if validators.url(selected):
			linkurl = data.selectedText()

		if linkurl != '':
			menu.AddItem('Open Link in New Tab', self._menu_actions, 'newtab', linkurl, icon=Icon('newtab'))
			menu.AddItem('Open Link in New Tab and Switch', self._menu_actions, 'newtabswitch', linkurl, icon=Icon('newtab'))
			menu.AddSeparator()
			menu.AddItem('Copy Link Location', self._menu_actions, 'copyurl')
			menu.AddItem('Copy Link Text', self._menu_actions, 'copy', linktext)

		if selected:
			if data.CanCut:
				menu.AddItem('Cut', self._menu_actions, 'cut', selected)

			if data.CanCopy:
				menu.AddItem('Copy', self._menu_actions, 'copy', selected)

		if data.isContentEditable():
			if data.CanPaste:
				menu.AddItem('Paste', self._menu_actions, 'paste', clipboard.text('plain'))

			if data.CanUndo:
				menu.AddSeparator()
				menu.AddItem('Undo', self._menu_actions, 'undo')

			if data.CanRedo:
				menu.AddItem('Redo', self._menu_actions, 'Redo')

		if linkurl != '':
			menu.AddItem('Download Link', self._menu_actions, 'download', linkurl)

			if acctcount and linkispost and linkismasto:
				menu.AddSeparator()
				menu.AddItem('Reply to Linked Post', self._mastodon_action, 'reply', linkurl)
				menu.AddItem('Favorite Linked Post', self._mastodon_action, 'favourite', linkurl)
				menu.AddItem('Boost Linked Post', self._mastodon_action, 'reblog', linkurl)
				menu.AddItem('Bookmark Linked Post', self._mastodon_action, 'bookmark', linkurl)
				#menu.AddItem('Add Domain Block', self._mastodon_action, 'domainblock', data.linkUrl().host())

		if menu.MoreThan(0):
			menu.AddSeparator()

		if selected and search:
			menu.AddSeparator()
			menu.AddItem(f'Search with {search.name}', self._menu_actions, 'newtabswitch', search.url.format(q=selected))
			searchmenu = Menu(self, 'Search with...')

			for row in db.fetch('search', single=False):
				if row.keyword != search.keyword:
					searchmenu.AddItem(row.name, self._menu_actions, 'newtabswitch', row.url.format(q=selected))

			menu.addMenu(searchmenu)
			menu.AddSeparator()

		if mediaurl != '':
			menu.AddItem('Open Media in New Tab', self._menu_actions, 'newtab', mediaurl, icon=Icon('newtab'))
			menu.AddItem('Open Media in New Tab and Switch', self._menu_actions, 'newtabswitch', mediaurl, icon=Icon('newtab'))
			menu.AddSeparator()
			menu.AddItem('Copy Media Location', self._menu_actions, 'copy', mediaurl)
			menu.AddItem('Download Media', self._menu_actions, 'download', mediaurl)
			menu.AddSeparator()

		menu.AddItem('Inspect Page', self._menu_actions, 'inspect_page')
		menu.AddItem('Inspect Element', self._menu_actions, 'inspect')
		menu.AddSeparator()
		menu.AddItem('View Page Source', self._menu_actions, 'source')

		menu.aboutToHide.connect(menu.deleteLater) #not sure if needed
		menu.popup(cursor)


	def _menu_actions(self, menuaction, data=None):
		actions = {
			'copy': {'cmd': self.window.clipboard.setText, 'args': [data]},
			'copyurl': {'cmd': self.webpage.triggerAction, 'args': [QWebEnginePage.CopyLinkToClipboard]},
			'paste': {'cmd': self.webpage.triggerAction, 'args': [QWebEnginePage.Paste]},
			'cut': {'cmd': self.webpage.triggerAction, 'args': [QWebEnginePage.Cut]},
			'undo': {'cmd': self.webpage.triggerAction, 'args': [QWebEnginePage.Undo]},
			'redo': {'cmd': self.webpage.triggerAction, 'args': [QWebEnginePage.Redo]},
			'selectall': {'cmd': self.webpage.triggerAction, 'args': [QWebEnginePage.SelectAll]},
			'reload': {'cmd': self.webpage.triggerAction, 'args': [QWebEnginePage.Reload]},
			'reloadforce': {'cmd': self.webpage.triggerAction, 'args': [QWebEnginePage.ReloadAndBypassCache]},
			'source': {'cmd': self.webpage.triggerAction, 'args': [QWebEnginePage.ViewSource]},
			'newtab': {'cmd': self.window.NewWebTab, 'args': [data], 'kwargs': {'nexttab': True}},
			'newtabswitch': {'cmd': self.window.NewWebTab, 'args': [data], 'kwargs': {'switch': True, 'nexttab': True}},
			'widgettab': {'cmd': self.window.NewWidgetTab, 'args': [data]},
			'download': {'cmd': self.webpage.DownloadUrl, 'args': [data]},
			'inspect': {'cmd': self.RunInspector, 'args': [True]},
			'inspect_page': {'cmd': self.RunInspector}
		}

		action = actions.get(menuaction)

		if not action:
			logging.error('invalid context menu action:', menuaction)
			return

		action['cmd'](*action.get('args', []), **action.get('kwargs', {}))


	def _mastodon_action(self, action, url):
		if action == 'domainblock':
			account = get.mastoaccount(default=True)
			self.window.NewWebTab(f'https://{account.domain}/admin/domain_blocks/new?_domain={url}')
		if action == 'reply':
			dialogs.Toot(self.window, self.window.mastodon, reply=url).show()

		if action in ['favourite', 'reblog', 'bookmark']:
			self.window.mastodon.PostAction(action, url, get.config('defaccount'))


	## Put the link in the status bar
	def _link_hover(self, url):
		self.statusurl = url

		label = self.window.ui.statusurl
		text = QFontMetrics(label.font()).elidedText(url, Qt.TextElideMode.ElideRight, label.width() - 10)
		label.setText(text)


	def _handle_js_error(self, level, message, line, sourceid):
		if get.config('jsdebug', None):
			logging.debug('js:', message)


	## Does nothing right now, but will be used later
	def CloseTab(self):
		pass


	def Fullscreen(self, gofull):
		fullscreen = self.window.FullscreenState()

		if gofull:
			self.window.ui.navbar.hide()
			self.window.tabbar.hide()
			self.window.ui.statusbar.hide()

			if fullscreen:
				self.isFullscreen = True

			else:
				self.isFullscreen = False
				self.window.Fullscreen(True)

		else:
			self.window.ui.navbar.show()
			self.window.tabbar.show()
			self.window.ui.statusbar.show()

			if not self.isFullscreen:
				self.window.Fullscreen(False)


	def RunInspector(self, element=False):
		if not self.devpage.view:
			self.webpage.setDevToolsPage(self.devpage.Setup(self.url().toString()))

		if element:
			self.webpage.triggerAction(QWebEnginePage.InspectElement)

		self.devpage.show()
		self.devpage.activateWindow()


	def Search(self, *args, reverse=False, sensitive=False):
		flags = QWebEnginePage.FindFlags(0)

		if len(args) == 0:
			self.page().findText('')
			return

		if sensitive:
			flags |= QWebEnginePage.FindCaseSensitively

		if reverse:
			flags |= QWebEnginePage.FindBackward

		self.page().findText(' '.join(args), flags)


	def LoadUrl(self, url):
		if not url:
			return

		if url == 'about:source':
			self.webpage.setUrl(QUrl('https://git.barkshark.xyz/izaliamae/qtweb'))
			return

		if any(map(url.startswith, ['chrome:', 'view-source:', 'dev-tools:'])):
			self.webpage.setUrl(QUrl(url))
			return

		newurl = 'https://' + url if not any(map(url.startswith, ['file://', 'ftp://', 'http://', 'https://', var.local])) else url
		parsed = urlparse(url)

		## Search instead of loading the url
		if not validators.url(newurl) and not url.startswith(var.local):
			split = url.split()
			keyword = split[0]
			row = get.search(keyword)

			if not row:
				return

			query = quote_plus(' '.join(split[1:]) if keyword == row.keyword else url)

			self.webpage.setUrl(QUrl(row.url.format(q=query)))
			return

		## redirect twitter links to nitter. will allow specifying the nitter instance later
		twitter_paths = any(map(parsed.path.startswith, ['/settings', '/home', '/login', '/logout', '/explore', '/sessions', '/account']))
		twitter_files = any(map(parsed.path.endswith, ['.js', '.jpeg', '.jpg', '.png', '.css']))

		if parsed.netloc == 'twitter.com' and get.config('nitter', False) and parsed.path != '/' and not twitter_paths and not twitter_files:
			nitterurl = get.config('nitterurl', default.nitter)
			newurl = nitterurl + parsed.path

		self.webpage.setUrl(QUrl(newurl))
		self.setFocus()


class SearchBar(QFrame):
	def __init__(self, parent):
		super().__init__(parent)
		self.webview = parent
		self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)

		layout = QHBoxLayout()
		layout.setContentsMargins(0, 6, 0, 0)
		layout.setSpacing(4)
		#layout.setSizeConstraint(QLayout.SetFixedSize)
		self.setLayout(layout)

		previous = QPushButton('Previous', self)
		next = QPushButton('Next', self)
		self.search = QLineEdit(self)
		self.sensitive = QCheckBox('Case Sensitive', self)
		reset = QPushButton('Reset', self)
		close = QPushButton(self)
		close.setFlat(True)
		close.setIcon(Icon('close'))

		for widget in [previous, next, self.search, self.sensitive, reset, close]:
			self.layout().addWidget(widget)

		self.search.installEventFilter(FindEsc(self))
		self.hide()

		self.search.returnPressed.connect(self.Search)
		next.clicked.connect(self.Search)
		previous.clicked.connect(lambda *args: self.Search('previous'))
		reset.clicked.connect(lambda *args: self.Search('reset'))
		close.clicked.connect(lambda *args: self.Search('toggle'))


	def Search(self, action=None, text=None):
		if text:
			self.search.setText(text)
		text = self.search.text()
		selected = self.webview.selectedText()
		sensitive = self.sensitive.isChecked()

		if action == 'previous':
			self.webview.Search(text, reverse=True, sensitive=sensitive)

		elif action == 'reset':
			self.webview.Search()
			self.sensitive.setChecked(False)
			self.search.setText('')

		elif action == 'toggle':
			self.setHidden(not self.isHidden())

			if not self.isHidden():
				## Focus the search box and set the text to the selected text if any
				self.search.setFocus()
				self.search.setText(selected)

				## If there is any selected text, start searching
				if selected != '':
					self.Search()

			else:
				## Reset the search and focus the window
				self.Search('reset')
				self.webview.setFocus()

		else:
			self.webview.Search(text, sensitive=sensitive)


class FindEsc(QObject):
	def __init__(self, parent):
		super().__init__(parent)
		self.webview = parent

	def eventFilter(self, obj, event):
		if event.type() == QEvent.KeyPress and event.key() == 16777216:
			self.webview.Search('toggle')
			self.webview.setFocus()

		return QObject.eventFilter(self, obj, event)


class WebEnginePage(QWebEnginePage):
	def __init__(self, profile, parent):
		super().__init__(profile, parent)
		self.webview = parent


	def certificateError(self, certerror):
		host = certerror.url().host()
		print(certerror.url())

		return True if host in ['localhost', '127.0.0.1'] or host.startswith('192.168.') else False


	def DownloadUrl(self, url):
		self.webview.webpage.download(QUrl(url))


	def chooseFile(self, _frame, suggested_file):
		filename, _ = QFileDialog.getOpenFileName(None, None, suggested_file)
		return filename


	def javaScriptConsoleMessage(self, level, message, line, sourceid):
		if get.config('jsdebug', False):
			logging.verbose('js:', message)


	def createWindow(self, wintype):
		options = {'nexttab': True}
		wintypes = {
			QWebEnginePage.WebBrowserTab: 'new tab',
			QWebEnginePage.WebBrowserWindow: 'new window',
			QWebEnginePage.WebBrowserBackgroundTab: 'new background tab',
			QWebEnginePage.WebDialog: 'new dialog'
		}

		typename = wintypes.get(wintype, f'Invalid window type: {type(wintype)}')

		if wintype != QWebEnginePage.WebDialog:
			options['switch'] = True

		return self.webview.window.NewWebTab('newtab' ,**options).webpage


	def triggerAction(self, action, checked=False):
		if action == QWebEnginePage.OpenLinkInNewWindow:
			self.createWindow(QWebEnginePage.WebBrowserWindow)
			return

		return super(WebEnginePage, self).triggerAction(action, checked)


	def acceptNavigationRequest(self, url, Type, isMainFrame):
		parsed = urlparse(url.toString())

		types = {
			QWebEnginePage.NavigationTypeLinkClicked: 'link clicked',
			QWebEnginePage.NavigationTypeTyped: 'loaded url',
			QWebEnginePage.NavigationTypeFormSubmitted: 'form submission',
			QWebEnginePage.NavigationTypeBackForward: 'back/forward action',
			QWebEnginePage.NavigationTypeReload: 'reloaded',
			QWebEnginePage.NavigationTypeOther: 'other type'
		}

		if not debstable:
			types[QWebEnginePage.NavigationTypeRedirect] = 'redirected'

		typename = types.get(Type, 'Invalid nav type')

		return QWebEnginePage.acceptNavigationRequest(self, url, Type, isMainFrame)


class DevPage(QMainWindow):
	def __init__(self, profile, parent):
		super().__init__()
		self.profile = profile
		self.window = parent
		self.view = None
		self.page = None
		#self.setGeometry(100, 100, 800, 600)
		self.resize(800, 600)

	def Setup(self, url):
		self.view = QWebEngineView(self.window)
		self.page = DevWebEnginePage(self.profile, self.window)
		self.view.setPage(self.page)
		self.setCentralWidget(self.view)
		self.setWindowTitle(f'QtWeb Inspector: {url}')
		self.show()
		self.activateWindow()

		return self.page


	def closeEvent(self, event=None):
		logging.debug('closing dev tools')
		self.view.deleteLater()
		self.page.deleteLater()
		self.view = None
		self.page = None

		if event:
			event.accept()


class DevWebEnginePage(QWebEnginePage):
	def __init__(self, profile, parent):
		super().__init__(profile, parent)
		self.window = parent


	def acceptNavigationRequest(self, url, Type, isMainFrame):
		if url.scheme() in ['chrome-devtools', 'devtools']:
			return QWebEnginePage.acceptNavigationRequest(self, url, Type, isMainFrame)

		else:
			options = {'nexttab': True, 'switch': True}
			self.window.NewWebTab(url.toString(), **options)
			return False


	def createWindow(self, wintype):
		options = {'nexttab': True, 'switch': True}
		return self.window.NewWebTab(**options).webpage
